import numpy as np
import pytest

import RATapi
import RATapi.classlist
import RATapi.outputs
import RATapi.rat_core


@pytest.fixture
def input_project():
    """A cut-down version of the input Project object for a reflectivity calculation set out in
    "DSPC_standard_layers.py".
    """
    project = RATapi.Project(
        name="original_dspc_bilayer",
        calculation="non polarised",
        model="standard layers",
        geometry="substrate/liquid",
        absorption=False,
    )

    # Set up the relevant parameters
    project.parameters.append(
        name="Oxide Thickness",
        min=5.0,
        value=19.54,
        max=60.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Oxide SLD",
        min=3.39e-06,
        value=3.39e-06,
        max=3.41e-06,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Tails Thickness",
        min=15.0,
        value=22.66,
        max=40.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Tails SLD",
        min=-5e-07,
        value=-4.01e-07,
        max=-3e-07,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Tails Hydration",
        min=1.0,
        value=5.252,
        max=50.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Roughness",
        min=1.0,
        value=5.64,
        max=15.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="CW Thickness",
        min=10.0,
        value=17.12,
        max=28.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="CW SLD",
        min=0.0,
        value=0.0,
        max=1e-09,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Heads Thickness",
        min=5.0,
        value=8.56,
        max=17.0,
        fit=True,
        prior_type="gaussian",
        mu=10.0,
        sigma=2.0,
    )
    project.parameters.append(
        name="SAM Heads SLD",
        min=1.0e-07,
        value=1.75e-06,
        max=2.0e-06,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="SAM Heads Hydration",
        min=10.0,
        value=45.45,
        max=50.0,
        fit=True,
        prior_type="uniform",
        mu=30.0,
        sigma=3.0,
    )
    project.parameters.append(
        name="Bilayer Heads Thickness",
        min=7.0,
        value=10.7,
        max=17.0,
        fit=True,
        prior_type="gaussian",
        mu=10.0,
        sigma=2.0,
    )
    project.parameters.append(
        name="Bilayer Heads SLD",
        min=5.0e-07,
        value=1.47e-06,
        max=1.5e-06,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Bilayer Roughness",
        min=2.0,
        value=6.014,
        max=15.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Bilayer Tails Thickness",
        min=14.0,
        value=17.82,
        max=22.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Bilayer Tails SLD",
        min=-5.0e-07,
        value=-4.61e-07,
        max=0.0,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Bilayer Tails Hydration",
        min=10.0,
        value=17.64,
        max=50.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Bilayer Heads Hydration",
        min=10.0,
        value=36.15,
        max=50.0,
        fit=True,
        prior_type="gaussian",
        mu=30.0,
        sigma=3.0,
    )
    project.parameters.append(
        name="CW Hydration",
        min=99.9,
        value=100.0,
        max=100.0,
        fit=False,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )
    project.parameters.append(
        name="Oxide Hydration",
        min=0.0,
        value=23.61,
        max=60.0,
        fit=True,
        prior_type="uniform",
        mu=0.0,
        sigma=np.inf,
    )

    project.parameters.set_fields(0, max=10)

    del project.bulk_in[0]
    project.bulk_in.append(name="Silicon", min=2.0e-06, value=2.073e-06, max=2.1e-06, fit=False)

    del project.bulk_out[0]
    project.bulk_out.append(name="D2O", min=5.50e-06, value=5.98e-06, max=6.4e-06, fit=True)
    project.bulk_out.append(name="SMW", min=1.0e-06, value=2.21e-06, max=4.99e-06, fit=True)

    del project.scalefactors[0]
    project.scalefactors.append(name="Scalefactor 1", min=0.05, value=0.10, max=0.2, fit=False)
    project.scalefactors.append(name="Scalefactor 2", min=0.05, value=0.15, max=0.2, fit=False)

    del project.backgrounds[0]
    del project.background_parameters[0]
    project.background_parameters.append(
        name="Background parameter D2O",
        min=5.0e-10,
        value=2.23e-06,
        max=7.0e-06,
        fit=True,
    )
    project.background_parameters.append(
        name="Background parameter SMW",
        min=1.0e-10,
        value=3.38e-06,
        max=4.99e-06,
        fit=True,
    )

    return project


@pytest.fixture
def reflectivity_calculation_output_results():
    """The C++ results object for a reflectivity calculation of the project set out in "DSPC_standard_layers.py"."""
    results = RATapi.rat_core.OutputResult()
    results.reflectivity = [
        np.array(
            [
                [1.14030000e-02, 1.00000223e00],
                [1.38610000e-02, 1.00000223e00],
                [1.68480000e-02, 6.42692703e-02],
                [2.04790000e-02, 1.69530247e-02],
                [2.48920000e-02, 6.18815370e-03],
                [3.02560000e-02, 2.99474163e-03],
                [3.67770000e-02, 1.73558265e-03],
                [4.47020000e-02, 1.03290554e-03],
                [5.43360000e-02, 6.45645642e-04],
                [6.60450000e-02, 5.57949197e-04],
                [8.02790000e-02, 5.58974858e-04],
                [9.75790000e-02, 3.49803750e-04],
                [1.18610000e-01, 8.89485853e-05],
                [1.44170000e-01, 4.04713900e-05],
                [1.75240000e-01, 6.98590391e-06],
                [2.13000000e-01, 3.60410795e-06],
                [2.58910000e-01, 2.41835671e-06],
                [3.14700000e-01, 2.26496635e-06],
                [3.82520000e-01, 2.26396064e-06],
                [4.64960000e-01, 2.24609517e-06],
                [5.65160000e-01, 2.23162115e-06],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.53247324e-02],
                [1.38610000e-02, 9.89602164e-03],
                [1.68480000e-02, 6.26868663e-03],
                [2.04790000e-02, 3.83918877e-03],
                [2.48920000e-02, 2.21818017e-03],
                [3.02560000e-02, 1.15473774e-03],
                [3.67770000e-02, 4.92469653e-04],
                [4.47020000e-02, 1.37287345e-04],
                [5.43360000e-02, 2.28561535e-05],
                [6.60450000e-02, 6.63620782e-05],
                [8.02790000e-02, 1.36790069e-04],
                [9.75790000e-02, 1.15354200e-04],
                [1.18610000e-01, 3.73063683e-05],
                [1.44170000e-01, 1.21263573e-05],
                [1.75240000e-01, 6.59103452e-06],
                [2.13000000e-01, 4.38321947e-06],
                [2.58910000e-01, 3.40797246e-06],
                [3.14700000e-01, 3.41625063e-06],
                [3.82520000e-01, 3.40076933e-06],
                [4.64960000e-01, 3.38892224e-06],
                [5.65160000e-01, 3.38103236e-06],
            ],
        ),
    ]
    results.simulation = [
        np.array(
            [
                [1.14030000e-02, 1.00000223e00],
                [1.38610000e-02, 1.00000223e00],
                [1.68480000e-02, 6.42692703e-02],
                [2.04790000e-02, 1.69530247e-02],
                [2.48920000e-02, 6.18815370e-03],
                [3.02560000e-02, 2.99474163e-03],
                [3.67770000e-02, 1.73558265e-03],
                [4.47020000e-02, 1.03290554e-03],
                [5.43360000e-02, 6.45645642e-04],
                [6.60450000e-02, 5.57949197e-04],
                [8.02790000e-02, 5.58974858e-04],
                [9.75790000e-02, 3.49803750e-04],
                [1.18610000e-01, 8.89485853e-05],
                [1.44170000e-01, 4.04713900e-05],
                [1.75240000e-01, 6.98590391e-06],
                [2.13000000e-01, 3.60410795e-06],
                [2.58910000e-01, 2.41835671e-06],
                [3.14700000e-01, 2.26496635e-06],
                [3.82520000e-01, 2.26396064e-06],
                [4.64960000e-01, 2.24609517e-06],
                [5.65160000e-01, 2.23162115e-06],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.53247324e-02],
                [1.38610000e-02, 9.89602164e-03],
                [1.68480000e-02, 6.26868663e-03],
                [2.04790000e-02, 3.83918877e-03],
                [2.48920000e-02, 2.21818017e-03],
                [3.02560000e-02, 1.15473774e-03],
                [3.67770000e-02, 4.92469653e-04],
                [4.47020000e-02, 1.37287345e-04],
                [5.43360000e-02, 2.28561535e-05],
                [6.60450000e-02, 6.63620782e-05],
                [8.02790000e-02, 1.36790069e-04],
                [9.75790000e-02, 1.15354200e-04],
                [1.18610000e-01, 3.73063683e-05],
                [1.44170000e-01, 1.21263573e-05],
                [1.75240000e-01, 6.59103452e-06],
                [2.13000000e-01, 4.38321947e-06],
                [2.58910000e-01, 3.40797246e-06],
                [3.14700000e-01, 3.41625063e-06],
                [3.82520000e-01, 3.40076933e-06],
                [4.64960000e-01, 3.38892224e-06],
                [5.65160000e-01, 3.38103236e-06],
            ],
        ),
    ]
    results.shiftedData = [
        np.array(
            [
                [1.1403e-02, 1.0063e00, 1.9003e-02],
                [1.3861e-02, 9.0118e-01, 1.1774e-02],
                [1.6848e-02, 7.0455e-02, 1.3083e-03],
                [2.0479e-02, 1.7544e-02, 5.1254e-04],
                [2.4892e-02, 6.4257e-03, 2.6236e-04],
                [3.0256e-02, 2.7746e-03, 8.5758e-05],
                [3.6777e-02, 1.8591e-03, 4.9391e-05],
                [4.4702e-02, 1.1002e-03, 3.2644e-05],
                [5.4336e-02, 6.6691e-04, 2.1365e-05],
                [6.6045e-02, 6.0729e-04, 1.1791e-05],
                [8.0279e-02, 5.8755e-04, 1.5569e-05],
                [9.7579e-02, 3.2700e-04, 6.5280e-06],
                [1.1861e-01, 7.8205e-05, 2.5881e-06],
                [1.4417e-01, 3.3455e-05, 1.5143e-06],
                [1.7524e-01, 4.9313e-06, 5.6663e-07],
                [2.1300e-01, 4.1948e-06, 6.8549e-07],
                [2.5891e-01, 3.9863e-06, 8.2061e-07],
                [3.1470e-01, 2.0861e-06, 6.1379e-07],
                [3.8252e-01, 2.1154e-06, 6.3084e-07],
                [4.6496e-01, 1.9906e-06, 6.0793e-07],
                [5.6516e-01, 2.3816e-06, 7.0610e-07],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.20493333e-02, 7.32200000e-04],
                [1.38610000e-02, 6.71800000e-03, 3.71853333e-04],
                [1.68480000e-02, 4.10506667e-03, 2.23106667e-04],
                [2.04790000e-02, 2.43306667e-03, 1.46873333e-04],
                [2.48920000e-02, 1.40940000e-03, 9.59200000e-05],
                [3.02560000e-02, 7.43400000e-04, 3.50226667e-05],
                [3.67770000e-02, 3.51466667e-04, 1.59573333e-05],
                [4.47020000e-02, 9.80666667e-05, 7.29466667e-06],
                [5.43360000e-02, 1.73500000e-05, 2.59200000e-06],
                [6.60450000e-02, 4.86286667e-05, 2.46453333e-06],
                [8.02790000e-02, 9.71733333e-05, 4.68166667e-06],
                [9.75790000e-02, 7.06533333e-05, 2.32180000e-06],
                [1.18610000e-01, 1.93046667e-05, 1.00813333e-06],
                [1.44170000e-01, 6.61193333e-06, 5.29773333e-07],
                [1.75240000e-01, 3.23726667e-06, 3.65933333e-07],
                [2.13000000e-01, 3.29920000e-06, 4.74273333e-07],
                [2.58910000e-01, 1.71180000e-06, 4.24720000e-07],
                [3.14700000e-01, 2.24020000e-06, 5.08946667e-07],
                [3.82520000e-01, 2.73306667e-06, 5.71513333e-07],
                [4.64960000e-01, 2.36153333e-06, 5.24806667e-07],
                [5.65160000e-01, 2.17460000e-06, 5.31346667e-07],
            ],
        ),
    ]
    results.layerSlds = [
        [
            np.array(
                [
                    [1.954000e01, 4.001499e-06, 3.000000e00],
                    [2.266000e01, -6.586988e-08, 3.000000e00],
                    [8.560000e00, 3.672535e-06, 5.640000e00],
                    [1.712000e01, 5.980000e-06, 5.640000e00],
                    [1.070000e01, 3.100365e-06, 6.014000e00],
                    [1.782000e01, 6.751924e-07, 6.014000e00],
                    [1.782000e01, 6.751924e-07, 6.014000e00],
                    [1.070000e01, 3.100365e-06, 6.014000e00],
                ],
            ),
        ],
        [
            np.array(
                [
                    [1.9540000e01, 3.1114020e-06, 3.0000000e00],
                    [2.2660000e01, -2.6387028e-07, 3.0000000e00],
                    [8.5600000e00, 1.9590700e-06, 5.6400000e00],
                    [1.7120000e01, 2.2100000e-06, 5.6400000e00],
                    [1.0700000e01, 1.7375100e-06, 6.0140000e00],
                    [1.7820000e01, 1.0164400e-08, 6.0140000e00],
                    [1.7820000e01, 1.0164400e-08, 6.0140000e00],
                    [1.0700000e01, 1.7375100e-06, 6.0140000e00],
                ],
            ),
        ],
    ]
    results.sldProfiles = [
        [
            np.array(
                [
                    [0.00000000e00, 2.07300000e-06],
                    [1.10000000e01, 2.07300000e-06],
                    [2.20000000e01, 2.07300000e-06],
                    [3.30000000e01, 2.07300001e-06],
                    [4.40000000e01, 2.11687361e-06],
                    [5.50000000e01, 3.90933280e-06],
                    [6.60000000e01, 3.51748792e-06],
                    [7.70000000e01, -2.64615370e-08],
                    [8.80000000e01, 8.14665196e-07],
                    [9.90000000e01, 4.11508879e-06],
                    [1.10000000e02, 5.58392432e-06],
                    [1.21000000e02, 3.71785214e-06],
                    [1.32000000e02, 1.39304140e-06],
                    [1.43000000e02, 6.95745588e-07],
                    [1.54000000e02, 7.84170141e-07],
                    [1.65000000e02, 2.15552213e-06],
                    [1.76000000e02, 4.68458332e-06],
                    [1.87000000e02, 5.91563638e-06],
                    [1.98000000e02, 5.97982117e-06],
                    [2.09000000e02, 5.97999998e-06],
                    [2.20000000e02, 5.98000000e-06],
                    [2.31000000e02, 5.98000000e-06],
                    [2.42000000e02, 5.98000000e-06],
                    [2.53000000e02, 5.98000000e-06],
                    [2.64000000e02, 5.98000000e-06],
                ],
            ),
            np.array(
                [
                    [0.00000000e00, 2.07300000e-06],
                    [1.10000000e01, 2.07300000e-06],
                    [2.20000000e01, 2.07300000e-06],
                    [3.30000000e01, 2.07300001e-06],
                    [4.40000000e01, 2.09662378e-06],
                    [5.50000000e01, 3.06177428e-06],
                    [6.60000000e01, 2.70974796e-06],
                    [7.70000000e01, -2.34283042e-07],
                    [8.80000000e01, 2.46446429e-07],
                    [9.90000000e01, 1.80004279e-06],
                    [1.10000000e02, 2.14886270e-06],
                    [1.21000000e02, 1.70090192e-06],
                    [1.32000000e02, 5.06554250e-07],
                    [1.43000000e02, 2.47801381e-08],
                    [1.54000000e02, 8.73866316e-08],
                    [1.65000000e02, 9.86362863e-07],
                    [1.76000000e02, 1.96411923e-06],
                    [1.87000000e02, 2.19933821e-06],
                    [1.98000000e02, 2.20997064e-06],
                    [2.09000000e02, 2.21000000e-06],
                    [2.20000000e02, 2.21000000e-06],
                    [2.31000000e02, 2.21000000e-06],
                    [2.42000000e02, 2.21000000e-06],
                    [2.53000000e02, 2.21000000e-06],
                    [2.64000000e02, 2.21000000e-06],
                ],
            ),
        ],
    ]
    results.resampledLayers = [[np.array([[0.0, 0.0, 0.0]])], [np.array([[0.0, 0.0, 0.0]])]]
    results.calculationResults = RATapi.rat_core.Calculation()
    results.calculationResults.chiValues = np.array([202.83057377, 1641.4024969])
    results.calculationResults.sumChi = 1844.2330706690975
    results.contrastParams = RATapi.rat_core.ContrastParams()
    results.contrastParams.backgroundParams = np.array([2.23e-06, 3.38e-06])
    results.contrastParams.scalefactors = np.array([0.1, 0.15])
    results.contrastParams.bulkIn = np.array([2.073e-06, 2.073e-06])
    results.contrastParams.bulkOut = np.array([5.98e-06, 2.21e-06])
    results.contrastParams.resolutionParams = np.array([0.03, 0.03])
    results.contrastParams.subRoughs = np.array([3.0, 3.0])
    results.contrastParams.resample = np.array([0.0, 0.0])
    results.fitParams = np.array(
        [
            3.000e00,
            1.954e01,
            2.266e01,
            5.252e00,
            5.640e00,
            1.712e01,
            8.560e00,
            4.545e01,
            1.070e01,
            6.014e00,
            1.782e01,
            1.764e01,
            3.615e01,
            2.361e01,
            2.230e-06,
            3.380e-06,
            5.980e-06,
            2.210e-06,
        ],
    )
    results.fitNames = [
        "Substrate Roughness",
        "Oxide Thickness",
        "SAM Tails Thickness",
        "SAM Tails Hydration",
        "SAM Roughness",
        "CW Thickness",
        "SAM Heads Thickness",
        "SAM Heads Hydration",
        "Bilayer Heads Thickness",
        "Bilayer Roughness",
        "Bilayer Tails Thickness",
        "Bilayer Tails Hydration",
        "Bilayer Heads Hydration",
        "Oxide Hydration",
        "Background parameter D2O",
        "Background parameter SMW",
        "D2O",
        "SMW",
    ]

    return results


@pytest.fixture
def reflectivity_calculation_results():
    """The python results object for a reflectivity calculation of the project set out in "DSPC_standard_layers.py"."""
    return RATapi.outputs.Results(
        reflectivity=[
            np.array(
                [
                    [1.14030000e-02, 1.00000223e00],
                    [1.38610000e-02, 1.00000223e00],
                    [1.68480000e-02, 6.42692703e-02],
                    [2.04790000e-02, 1.69530247e-02],
                    [2.48920000e-02, 6.18815370e-03],
                    [3.02560000e-02, 2.99474163e-03],
                    [3.67770000e-02, 1.73558265e-03],
                    [4.47020000e-02, 1.03290554e-03],
                    [5.43360000e-02, 6.45645642e-04],
                    [6.60450000e-02, 5.57949197e-04],
                    [8.02790000e-02, 5.58974858e-04],
                    [9.75790000e-02, 3.49803750e-04],
                    [1.18610000e-01, 8.89485853e-05],
                    [1.44170000e-01, 4.04713900e-05],
                    [1.75240000e-01, 6.98590391e-06],
                    [2.13000000e-01, 3.60410795e-06],
                    [2.58910000e-01, 2.41835671e-06],
                    [3.14700000e-01, 2.26496635e-06],
                    [3.82520000e-01, 2.26396064e-06],
                    [4.64960000e-01, 2.24609517e-06],
                    [5.65160000e-01, 2.23162115e-06],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.53247324e-02],
                    [1.38610000e-02, 9.89602164e-03],
                    [1.68480000e-02, 6.26868663e-03],
                    [2.04790000e-02, 3.83918877e-03],
                    [2.48920000e-02, 2.21818017e-03],
                    [3.02560000e-02, 1.15473774e-03],
                    [3.67770000e-02, 4.92469653e-04],
                    [4.47020000e-02, 1.37287345e-04],
                    [5.43360000e-02, 2.28561535e-05],
                    [6.60450000e-02, 6.63620782e-05],
                    [8.02790000e-02, 1.36790069e-04],
                    [9.75790000e-02, 1.15354200e-04],
                    [1.18610000e-01, 3.73063683e-05],
                    [1.44170000e-01, 1.21263573e-05],
                    [1.75240000e-01, 6.59103452e-06],
                    [2.13000000e-01, 4.38321947e-06],
                    [2.58910000e-01, 3.40797246e-06],
                    [3.14700000e-01, 3.41625063e-06],
                    [3.82520000e-01, 3.40076933e-06],
                    [4.64960000e-01, 3.38892224e-06],
                    [5.65160000e-01, 3.38103236e-06],
                ],
            ),
        ],
        simulation=[
            np.array(
                [
                    [1.14030000e-02, 1.00000223e00],
                    [1.38610000e-02, 1.00000223e00],
                    [1.68480000e-02, 6.42692703e-02],
                    [2.04790000e-02, 1.69530247e-02],
                    [2.48920000e-02, 6.18815370e-03],
                    [3.02560000e-02, 2.99474163e-03],
                    [3.67770000e-02, 1.73558265e-03],
                    [4.47020000e-02, 1.03290554e-03],
                    [5.43360000e-02, 6.45645642e-04],
                    [6.60450000e-02, 5.57949197e-04],
                    [8.02790000e-02, 5.58974858e-04],
                    [9.75790000e-02, 3.49803750e-04],
                    [1.18610000e-01, 8.89485853e-05],
                    [1.44170000e-01, 4.04713900e-05],
                    [1.75240000e-01, 6.98590391e-06],
                    [2.13000000e-01, 3.60410795e-06],
                    [2.58910000e-01, 2.41835671e-06],
                    [3.14700000e-01, 2.26496635e-06],
                    [3.82520000e-01, 2.26396064e-06],
                    [4.64960000e-01, 2.24609517e-06],
                    [5.65160000e-01, 2.23162115e-06],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.53247324e-02],
                    [1.38610000e-02, 9.89602164e-03],
                    [1.68480000e-02, 6.26868663e-03],
                    [2.04790000e-02, 3.83918877e-03],
                    [2.48920000e-02, 2.21818017e-03],
                    [3.02560000e-02, 1.15473774e-03],
                    [3.67770000e-02, 4.92469653e-04],
                    [4.47020000e-02, 1.37287345e-04],
                    [5.43360000e-02, 2.28561535e-05],
                    [6.60450000e-02, 6.63620782e-05],
                    [8.02790000e-02, 1.36790069e-04],
                    [9.75790000e-02, 1.15354200e-04],
                    [1.18610000e-01, 3.73063683e-05],
                    [1.44170000e-01, 1.21263573e-05],
                    [1.75240000e-01, 6.59103452e-06],
                    [2.13000000e-01, 4.38321947e-06],
                    [2.58910000e-01, 3.40797246e-06],
                    [3.14700000e-01, 3.41625063e-06],
                    [3.82520000e-01, 3.40076933e-06],
                    [4.64960000e-01, 3.38892224e-06],
                    [5.65160000e-01, 3.38103236e-06],
                ],
            ),
        ],
        shiftedData=[
            np.array(
                [
                    [1.1403e-02, 1.0063e00, 1.9003e-02],
                    [1.3861e-02, 9.0118e-01, 1.1774e-02],
                    [1.6848e-02, 7.0455e-02, 1.3083e-03],
                    [2.0479e-02, 1.7544e-02, 5.1254e-04],
                    [2.4892e-02, 6.4257e-03, 2.6236e-04],
                    [3.0256e-02, 2.7746e-03, 8.5758e-05],
                    [3.6777e-02, 1.8591e-03, 4.9391e-05],
                    [4.4702e-02, 1.1002e-03, 3.2644e-05],
                    [5.4336e-02, 6.6691e-04, 2.1365e-05],
                    [6.6045e-02, 6.0729e-04, 1.1791e-05],
                    [8.0279e-02, 5.8755e-04, 1.5569e-05],
                    [9.7579e-02, 3.2700e-04, 6.5280e-06],
                    [1.1861e-01, 7.8205e-05, 2.5881e-06],
                    [1.4417e-01, 3.3455e-05, 1.5143e-06],
                    [1.7524e-01, 4.9313e-06, 5.6663e-07],
                    [2.1300e-01, 4.1948e-06, 6.8549e-07],
                    [2.5891e-01, 3.9863e-06, 8.2061e-07],
                    [3.1470e-01, 2.0861e-06, 6.1379e-07],
                    [3.8252e-01, 2.1154e-06, 6.3084e-07],
                    [4.6496e-01, 1.9906e-06, 6.0793e-07],
                    [5.6516e-01, 2.3816e-06, 7.0610e-07],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.20493333e-02, 7.32200000e-04],
                    [1.38610000e-02, 6.71800000e-03, 3.71853333e-04],
                    [1.68480000e-02, 4.10506667e-03, 2.23106667e-04],
                    [2.04790000e-02, 2.43306667e-03, 1.46873333e-04],
                    [2.48920000e-02, 1.40940000e-03, 9.59200000e-05],
                    [3.02560000e-02, 7.43400000e-04, 3.50226667e-05],
                    [3.67770000e-02, 3.51466667e-04, 1.59573333e-05],
                    [4.47020000e-02, 9.80666667e-05, 7.29466667e-06],
                    [5.43360000e-02, 1.73500000e-05, 2.59200000e-06],
                    [6.60450000e-02, 4.86286667e-05, 2.46453333e-06],
                    [8.02790000e-02, 9.71733333e-05, 4.68166667e-06],
                    [9.75790000e-02, 7.06533333e-05, 2.32180000e-06],
                    [1.18610000e-01, 1.93046667e-05, 1.00813333e-06],
                    [1.44170000e-01, 6.61193333e-06, 5.29773333e-07],
                    [1.75240000e-01, 3.23726667e-06, 3.65933333e-07],
                    [2.13000000e-01, 3.29920000e-06, 4.74273333e-07],
                    [2.58910000e-01, 1.71180000e-06, 4.24720000e-07],
                    [3.14700000e-01, 2.24020000e-06, 5.08946667e-07],
                    [3.82520000e-01, 2.73306667e-06, 5.71513333e-07],
                    [4.64960000e-01, 2.36153333e-06, 5.24806667e-07],
                    [5.65160000e-01, 2.17460000e-06, 5.31346667e-07],
                ],
            ),
        ],
        layerSlds=[
            [
                np.array(
                    [
                        [1.954000e01, 4.001499e-06, 3.000000e00],
                        [2.266000e01, -6.586988e-08, 3.000000e00],
                        [8.560000e00, 3.672535e-06, 5.640000e00],
                        [1.712000e01, 5.980000e-06, 5.640000e00],
                        [1.070000e01, 3.100365e-06, 6.014000e00],
                        [1.782000e01, 6.751924e-07, 6.014000e00],
                        [1.782000e01, 6.751924e-07, 6.014000e00],
                        [1.070000e01, 3.100365e-06, 6.014000e00],
                    ],
                ),
            ],
            [
                np.array(
                    [
                        [1.9540000e01, 3.1114020e-06, 3.0000000e00],
                        [2.2660000e01, -2.6387028e-07, 3.0000000e00],
                        [8.5600000e00, 1.9590700e-06, 5.6400000e00],
                        [1.7120000e01, 2.2100000e-06, 5.6400000e00],
                        [1.0700000e01, 1.7375100e-06, 6.0140000e00],
                        [1.7820000e01, 1.0164400e-08, 6.0140000e00],
                        [1.7820000e01, 1.0164400e-08, 6.0140000e00],
                        [1.0700000e01, 1.7375100e-06, 6.0140000e00],
                    ],
                ),
            ],
        ],
        sldProfiles=[
            [
                np.array(
                    [
                        [0.00000000e00, 2.07300000e-06],
                        [1.10000000e01, 2.07300000e-06],
                        [2.20000000e01, 2.07300000e-06],
                        [3.30000000e01, 2.07300001e-06],
                        [4.40000000e01, 2.11687361e-06],
                        [5.50000000e01, 3.90933280e-06],
                        [6.60000000e01, 3.51748792e-06],
                        [7.70000000e01, -2.64615370e-08],
                        [8.80000000e01, 8.14665196e-07],
                        [9.90000000e01, 4.11508879e-06],
                        [1.10000000e02, 5.58392432e-06],
                        [1.21000000e02, 3.71785214e-06],
                        [1.32000000e02, 1.39304140e-06],
                        [1.43000000e02, 6.95745588e-07],
                        [1.54000000e02, 7.84170141e-07],
                        [1.65000000e02, 2.15552213e-06],
                        [1.76000000e02, 4.68458332e-06],
                        [1.87000000e02, 5.91563638e-06],
                        [1.98000000e02, 5.97982117e-06],
                        [2.09000000e02, 5.97999998e-06],
                        [2.20000000e02, 5.98000000e-06],
                        [2.31000000e02, 5.98000000e-06],
                        [2.42000000e02, 5.98000000e-06],
                        [2.53000000e02, 5.98000000e-06],
                        [2.64000000e02, 5.98000000e-06],
                    ],
                ),
                np.array(
                    [
                        [0.00000000e00, 2.07300000e-06],
                        [1.10000000e01, 2.07300000e-06],
                        [2.20000000e01, 2.07300000e-06],
                        [3.30000000e01, 2.07300001e-06],
                        [4.40000000e01, 2.09662378e-06],
                        [5.50000000e01, 3.06177428e-06],
                        [6.60000000e01, 2.70974796e-06],
                        [7.70000000e01, -2.34283042e-07],
                        [8.80000000e01, 2.46446429e-07],
                        [9.90000000e01, 1.80004279e-06],
                        [1.10000000e02, 2.14886270e-06],
                        [1.21000000e02, 1.70090192e-06],
                        [1.32000000e02, 5.06554250e-07],
                        [1.43000000e02, 2.47801381e-08],
                        [1.54000000e02, 8.73866316e-08],
                        [1.65000000e02, 9.86362863e-07],
                        [1.76000000e02, 1.96411923e-06],
                        [1.87000000e02, 2.19933821e-06],
                        [1.98000000e02, 2.20997064e-06],
                        [2.09000000e02, 2.21000000e-06],
                        [2.20000000e02, 2.21000000e-06],
                        [2.31000000e02, 2.21000000e-06],
                        [2.42000000e02, 2.21000000e-06],
                        [2.53000000e02, 2.21000000e-06],
                        [2.64000000e02, 2.21000000e-06],
                    ],
                ),
            ],
        ],
        resampledLayers=[[np.array([[0.0, 0.0, 0.0]])], [np.array([[0.0, 0.0, 0.0]])]],
        calculationResults=RATapi.outputs.CalculationResults(
            chiValues=np.array([202.83057377, 1641.4024969]),
            sumChi=1844.2330706690975,
        ),
        contrastParams=RATapi.outputs.ContrastParams(
            backgroundParams=np.array([2.23e-06, 3.38e-06]),
            scalefactors=np.array([0.1, 0.15]),
            bulkIn=np.array([2.073e-06, 2.073e-06]),
            bulkOut=np.array([5.98e-06, 2.21e-06]),
            resolutionParams=np.array([0.03, 0.03]),
            subRoughs=np.array([3.0, 3.0]),
            resample=np.array([0.0, 0.0]),
        ),
        fitParams=np.array(
            [
                3.000e00,
                1.954e01,
                2.266e01,
                5.252e00,
                5.640e00,
                1.712e01,
                8.560e00,
                4.545e01,
                1.070e01,
                6.014e00,
                1.782e01,
                1.764e01,
                3.615e01,
                2.361e01,
                2.230e-06,
                3.380e-06,
                5.980e-06,
                2.210e-06,
            ],
        ),
        fitNames=[
            "Substrate Roughness",
            "Oxide Thickness",
            "SAM Tails Thickness",
            "SAM Tails Hydration",
            "SAM Roughness",
            "CW Thickness",
            "SAM Heads Thickness",
            "SAM Heads Hydration",
            "Bilayer Heads Thickness",
            "Bilayer Roughness",
            "Bilayer Tails Thickness",
            "Bilayer Tails Hydration",
            "Bilayer Heads Hydration",
            "Oxide Hydration",
            "Background parameter D2O",
            "Background parameter SMW",
            "D2O",
            "SMW",
        ],
    )


@pytest.fixture
def dream_output_results():
    """The C++ results object for a Dream optimisation of the project set out in "DSPC_standard_layers.py".

    This optimisation used the parameters: nSamples=1, nChains=1. However, the calculationResults, contrastParams,
    and fitParams are taken from an optimisation with the parameters: nSamples=50000, nChains=10.
    """
    results = RATapi.rat_core.OutputResult()
    results.reflectivity = [
        np.array(
            [
                [1.14030000e-02, 1.45891447e01],
                [1.38610000e-02, 1.45739839e01],
                [1.68480000e-02, 1.45684725e01],
                [2.04790000e-02, 1.45658452e01],
                [2.48920000e-02, 1.45641513e01],
                [3.02560000e-02, 1.45628311e01],
                [3.67770000e-02, 1.45618510e01],
                [4.47020000e-02, 1.45613427e01],
                [5.43360000e-02, 1.45612940e01],
                [6.60450000e-02, 1.45613971e01],
                [8.02790000e-02, 1.45613678e01],
                [9.75790000e-02, 1.45612960e01],
                [1.18610000e-01, 1.45612799e01],
                [1.44170000e-01, 1.45612749e01],
                [1.75240000e-01, 1.45612726e01],
                [2.13000000e-01, 1.45612723e01],
                [2.58910000e-01, 1.45612724e01],
                [3.14700000e-01, 1.45612723e01],
                [3.82520000e-01, 1.45612723e01],
                [4.64960000e-01, 1.45612723e01],
                [5.65160000e-01, 1.45612723e01],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.00000316e00],
                [1.38610000e-02, 1.21238320e-01],
                [1.68480000e-02, 2.13433068e-02],
                [2.04790000e-02, 8.71162144e-03],
                [2.48920000e-02, 5.72807902e-03],
                [3.02560000e-02, 3.71827525e-03],
                [3.67770000e-02, 1.72067772e-03],
                [4.47020000e-02, 3.51471141e-04],
                [5.43360000e-02, 2.03975460e-05],
                [6.60450000e-02, 1.89221181e-04],
                [8.02790000e-02, 1.49280488e-04],
                [9.75790000e-02, 3.86439083e-05],
                [1.18610000e-01, 1.63612094e-05],
                [1.44170000e-01, 6.75770325e-06],
                [1.75240000e-01, 3.45955607e-06],
                [2.13000000e-01, 3.20880782e-06],
                [2.58910000e-01, 3.32068345e-06],
                [3.14700000e-01, 3.21671807e-06],
                [3.82520000e-01, 3.15612021e-06],
                [4.64960000e-01, 3.15645546e-06],
                [5.65160000e-01, 3.15511510e-06],
            ],
        ),
    ]
    results.simulation = [
        np.array(
            [
                [1.14030000e-02, 1.45891447e01],
                [1.38610000e-02, 1.45739839e01],
                [1.68480000e-02, 1.45684725e01],
                [2.04790000e-02, 1.45658452e01],
                [2.48920000e-02, 1.45641513e01],
                [3.02560000e-02, 1.45628311e01],
                [3.67770000e-02, 1.45618510e01],
                [4.47020000e-02, 1.45613427e01],
                [5.43360000e-02, 1.45612940e01],
                [6.60450000e-02, 1.45613971e01],
                [8.02790000e-02, 1.45613678e01],
                [9.75790000e-02, 1.45612960e01],
                [1.18610000e-01, 1.45612799e01],
                [1.44170000e-01, 1.45612749e01],
                [1.75240000e-01, 1.45612726e01],
                [2.13000000e-01, 1.45612723e01],
                [2.58910000e-01, 1.45612724e01],
                [3.14700000e-01, 1.45612723e01],
                [3.82520000e-01, 1.45612723e01],
                [4.64960000e-01, 1.45612723e01],
                [5.65160000e-01, 1.45612723e01],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.00000316e00],
                [1.38610000e-02, 1.21238320e-01],
                [1.68480000e-02, 2.13433068e-02],
                [2.04790000e-02, 8.71162144e-03],
                [2.48920000e-02, 5.72807902e-03],
                [3.02560000e-02, 3.71827525e-03],
                [3.67770000e-02, 1.72067772e-03],
                [4.47020000e-02, 3.51471141e-04],
                [5.43360000e-02, 2.03975460e-05],
                [6.60450000e-02, 1.89221181e-04],
                [8.02790000e-02, 1.49280488e-04],
                [9.75790000e-02, 3.86439083e-05],
                [1.18610000e-01, 1.63612094e-05],
                [1.44170000e-01, 6.75770325e-06],
                [1.75240000e-01, 3.45955607e-06],
                [2.13000000e-01, 3.20880782e-06],
                [2.58910000e-01, 3.32068345e-06],
                [3.14700000e-01, 3.21671807e-06],
                [3.82520000e-01, 3.15612021e-06],
                [4.64960000e-01, 3.15645546e-06],
                [5.65160000e-01, 3.15511510e-06],
            ],
        ),
    ]
    results.shiftedData = [
        np.array(
            [
                [1.1403e-02, 1.0063e00, 1.9003e-02],
                [1.3861e-02, 9.0118e-01, 1.1774e-02],
                [1.6848e-02, 7.0455e-02, 1.3083e-03],
                [2.0479e-02, 1.7544e-02, 5.1254e-04],
                [2.4892e-02, 6.4257e-03, 2.6236e-04],
                [3.0256e-02, 2.7746e-03, 8.5758e-05],
                [3.6777e-02, 1.8591e-03, 4.9391e-05],
                [4.4702e-02, 1.1002e-03, 3.2644e-05],
                [5.4336e-02, 6.6691e-04, 2.1365e-05],
                [6.6045e-02, 6.0729e-04, 1.1791e-05],
                [8.0279e-02, 5.8755e-04, 1.5569e-05],
                [9.7579e-02, 3.2700e-04, 6.5280e-06],
                [1.1861e-01, 7.8205e-05, 2.5881e-06],
                [1.4417e-01, 3.3455e-05, 1.5143e-06],
                [1.7524e-01, 4.9313e-06, 5.6663e-07],
                [2.1300e-01, 4.1948e-06, 6.8549e-07],
                [2.5891e-01, 3.9863e-06, 8.2061e-07],
                [3.1470e-01, 2.0861e-06, 6.1379e-07],
                [3.8252e-01, 2.1154e-06, 6.3084e-07],
                [4.6496e-01, 1.9906e-06, 6.0793e-07],
                [5.6516e-01, 2.3816e-06, 7.0610e-07],
            ],
        ),
        np.array(
            [
                [1.14030000e-02, 1.20493333e-02, 7.32200000e-04],
                [1.38610000e-02, 6.71800000e-03, 3.71853333e-04],
                [1.68480000e-02, 4.10506667e-03, 2.23106667e-04],
                [2.04790000e-02, 2.43306667e-03, 1.46873333e-04],
                [2.48920000e-02, 1.40940000e-03, 9.59200000e-05],
                [3.02560000e-02, 7.43400000e-04, 3.50226667e-05],
                [3.67770000e-02, 3.51466667e-04, 1.59573333e-05],
                [4.47020000e-02, 9.80666667e-05, 7.29466667e-06],
                [5.43360000e-02, 1.73500000e-05, 2.59200000e-06],
                [6.60450000e-02, 4.86286667e-05, 2.46453333e-06],
                [8.02790000e-02, 9.71733333e-05, 4.68166667e-06],
                [9.75790000e-02, 7.06533333e-05, 2.32180000e-06],
                [1.18610000e-01, 1.93046667e-05, 1.00813333e-06],
                [1.44170000e-01, 6.61193333e-06, 5.29773333e-07],
                [1.75240000e-01, 3.23726667e-06, 3.65933333e-07],
                [2.13000000e-01, 3.29920000e-06, 4.74273333e-07],
                [2.58910000e-01, 1.71180000e-06, 4.24720000e-07],
                [3.14700000e-01, 2.24020000e-06, 5.08946667e-07],
                [3.82520000e-01, 2.73306667e-06, 5.71513333e-07],
                [4.64960000e-01, 2.36153333e-06, 5.24806667e-07],
                [5.65160000e-01, 2.17460000e-06, 5.31346667e-07],
            ],
        ),
    ]
    results.layerSlds = [
        [
            np.array(
                [
                    [3.15755349e01, 3.35278238e-06, 4.16625659e00],
                    [3.61791464e01, 7.68327921e-07, 4.16625659e00],
                    [1.00488530e01, 2.06044530e-06, 2.78042232e01],
                    [1.08043784e01, 3.29384190e-06, 2.78042232e01],
                    [2.42251646e01, 2.35556998e-06, 1.55593097e01],
                    [1.49022278e01, 7.42138004e-07, 1.55593097e01],
                    [1.49022278e01, 7.42138004e-07, 1.55593097e01],
                    [2.42251646e01, 2.35556998e-06, 1.55593097e01],
                ],
            ),
        ],
        [
            np.array(
                [
                    [3.15755349e01, 4.11636356e-06, 4.16625659e00],
                    [3.61791464e01, 1.39268494e-06, 4.16625659e00],
                    [1.00488530e01, 2.45715680e-06, 2.78042232e01],
                    [1.08043784e01, 5.26668495e-06, 2.78042232e01],
                    [2.42251646e01, 3.31348777e-06, 1.55593097e01],
                    [1.49022278e01, 1.37428245e-06, 1.55593097e01],
                    [1.49022278e01, 1.37428245e-06, 1.55593097e01],
                    [2.42251646e01, 3.31348777e-06, 1.55593097e01],
                ],
            ),
        ],
    ]
    results.sldProfiles = [
        [
            np.array(
                [
                    [0.00000000e00, 2.07301741e-06],
                    [1.10000000e01, 2.07309604e-06],
                    [2.20000000e01, 2.07345780e-06],
                    [3.30000000e01, 2.07491687e-06],
                    [4.40000000e01, 2.17562259e-06],
                    [5.50000000e01, 3.22650495e-06],
                    [6.60000000e01, 3.40913848e-06],
                    [7.70000000e01, 3.13506253e-06],
                    [8.80000000e01, 1.20456289e-06],
                    [9.90000000e01, 1.27138002e-06],
                    [1.10000000e02, 1.56285495e-06],
                    [1.21000000e02, 1.84518709e-06],
                    [1.32000000e02, 2.00498449e-06],
                    [1.43000000e02, 1.96260665e-06],
                    [1.54000000e02, 1.71966879e-06],
                    [1.65000000e02, 1.39383267e-06],
                    [1.76000000e02, 1.22970195e-06],
                    [1.87000000e02, 1.41565526e-06],
                    [1.98000000e02, 1.88172657e-06],
                    [2.09000000e02, 2.40655867e-06],
                    [2.20000000e02, 2.83570420e-06],
                    [2.31000000e02, 3.11210082e-06],
                    [2.42000000e02, 3.24277802e-06],
                    [2.53000000e02, 3.28427452e-06],
                    [2.64000000e02, 3.29268847e-06],
                    [2.75000000e02, 3.29375418e-06],
                    [2.86000000e02, 3.29383773e-06],
                    [2.97000000e02, 3.29384178e-06],
                    [3.08000000e02, 3.29384190e-06],
                ],
            ),
            np.array(
                [
                    [0.00000000e00, 2.07301819e-06],
                    [1.10000000e01, 2.07310296e-06],
                    [2.20000000e01, 2.07350415e-06],
                    [3.30000000e01, 2.07518460e-06],
                    [4.40000000e01, 2.23394767e-06],
                    [5.50000000e01, 3.90646736e-06],
                    [6.60000000e01, 4.18619279e-06],
                    [7.70000000e01, 3.91675004e-06],
                    [8.80000000e01, 1.92451547e-06],
                    [9.90000000e01, 2.06988861e-06],
                    [1.10000000e02, 2.47614934e-06],
                    [1.21000000e02, 2.84889288e-06],
                    [1.32000000e02, 3.00529286e-06],
                    [1.43000000e02, 2.86310144e-06],
                    [1.54000000e02, 2.49967205e-06],
                    [1.65000000e02, 2.09673058e-06],
                    [1.76000000e02, 1.92431424e-06],
                    [1.87000000e02, 2.18829605e-06],
                    [1.98000000e02, 2.83032190e-06],
                    [2.09000000e02, 3.62584708e-06],
                    [2.20000000e02, 4.36871153e-06],
                    [2.31000000e02, 4.89806866e-06],
                    [2.42000000e02, 5.16145928e-06],
                    [2.53000000e02, 5.24684298e-06],
                    [2.64000000e02, 5.26428707e-06],
                    [2.75000000e02, 5.26650242e-06],
                    [2.86000000e02, 5.26667628e-06],
                    [2.97000000e02, 5.26668469e-06],
                    [3.08000000e02, 5.26668495e-06],
                ],
            ),
        ],
    ]
    results.resampledLayers = [[np.array([[0.0, 0.0, 0.0]])], [np.array([[0.0, 0.0, 0.0]])]]
    results.calculationResults = RATapi.rat_core.Calculation()
    results.calculationResults.chiValues = (np.array([4.6077885, 7.00028098]),)
    results.calculationResults.sumChi = 11.608069475997699
    results.contrastParams = RATapi.rat_core.ContrastParams()
    results.contrastParams.backgroundParams = np.array([2.37113128e-06, 1.99006694e-06])
    results.contrastParams.scalefactors = np.array([0.1, 0.15])
    results.contrastParams.bulkIn = np.array([2.073e-06, 2.073e-06])
    results.contrastParams.bulkOut = np.array([6.01489149e-06, 1.59371685e-06])
    results.contrastParams.resolutionParams = np.array([0.03, 0.03])
    results.contrastParams.subRoughs = np.array([6.19503045, 6.19503045])
    results.contrastParams.resample = np.array([0.0, 0.0])
    results.fitParams = np.array(
        [
            6.19503045e00,
            1.84420960e01,
            2.11039621e01,
            8.75538121e00,
            3.72292994e00,
            1.84624551e01,
            1.02316734e01,
            2.31156093e01,
            1.09906265e01,
            5.71005361e00,
            1.67933822e01,
            1.72009856e01,
            3.00260126e01,
            2.94448999e01,
            2.37113128e-06,
            1.99006694e-06,
            6.01489149e-06,
            1.59371685e-06,
        ],
    )
    results.fitNames = [
        "Substrate Roughness",
        "Oxide Thickness",
        "SAM Tails Thickness",
        "SAM Tails Hydration",
        "SAM Roughness",
        "CW Thickness",
        "SAM Heads Thickness",
        "SAM Heads Hydration",
        "Bilayer Heads Thickness",
        "Bilayer Roughness",
        "Bilayer Tails Thickness",
        "Bilayer Tails Hydration",
        "Bilayer Heads Hydration",
        "Oxide Hydration",
        "Background parameter D2O",
        "Background parameter SMW",
        "D2O",
        "SMW",
    ]

    return results


@pytest.fixture
def dream_bayes():
    """The C++ BayesResults object for a dream optimisation of the project set out in "DSPC_standard_layers.py".

    This optimisation used the parameters: nSamples=1, nChains=1.
    """
    bayes = RATapi.rat_core.BayesResults()
    bayes.predictionIntervals = RATapi.rat_core.PredictionIntervals()
    bayes.predictionIntervals.reflectivity = [
        np.array(
            [
                [
                    1.00000560e00,
                    7.07687612e-01,
                    2.08315160e-02,
                    2.40787966e-03,
                    2.17627660e-03,
                    2.54301700e-03,
                    1.76309827e-03,
                    6.15269679e-04,
                    3.43679710e-05,
                    7.93625275e-05,
                    1.27760549e-04,
                    3.35799941e-05,
                    6.75048751e-06,
                    7.64258431e-06,
                    6.05800395e-06,
                    5.60288298e-06,
                    5.60333677e-06,
                    5.60223341e-06,
                    5.60206667e-06,
                    5.60206314e-06,
                    5.60206314e-06,
                ],
                [
                    1.00000560e00,
                    7.07687612e-01,
                    2.08315160e-02,
                    2.40787966e-03,
                    2.17627660e-03,
                    2.54301700e-03,
                    1.76309827e-03,
                    6.15269679e-04,
                    3.43679710e-05,
                    7.93625275e-05,
                    1.27760549e-04,
                    3.35799941e-05,
                    6.75048751e-06,
                    7.64258431e-06,
                    6.05800395e-06,
                    5.60288298e-06,
                    5.60333677e-06,
                    5.60223341e-06,
                    5.60206667e-06,
                    5.60206314e-06,
                    5.60206314e-06,
                ],
                [
                    1.55201098e01,
                    1.53748894e01,
                    1.50402011e01,
                    1.50299478e01,
                    1.50290525e01,
                    1.50287935e01,
                    1.50282126e01,
                    1.50276044e01,
                    1.50273248e01,
                    1.50273185e01,
                    1.50272997e01,
                    1.50272498e01,
                    1.50272364e01,
                    1.50272347e01,
                    1.50272335e01,
                    1.50272330e01,
                    1.50272329e01,
                    1.50272328e01,
                    1.50272328e01,
                    1.50272328e01,
                    1.50272328e01,
                ],
                [
                    2.91397425e01,
                    2.91324974e01,
                    2.91281368e01,
                    2.91255472e01,
                    2.91240295e01,
                    2.91231836e01,
                    2.91227893e01,
                    2.91226872e01,
                    2.91226902e01,
                    2.91226359e01,
                    2.91225541e01,
                    2.91225457e01,
                    2.91225448e01,
                    2.91225407e01,
                    2.91225398e01,
                    2.91225393e01,
                    2.91225391e01,
                    2.91225390e01,
                    2.91225390e01,
                    2.91225389e01,
                    2.91225389e01,
                ],
                [
                    2.91397425e01,
                    2.91324974e01,
                    2.91281368e01,
                    2.91255472e01,
                    2.91240295e01,
                    2.91231836e01,
                    2.91227893e01,
                    2.91226872e01,
                    2.91226902e01,
                    2.91226359e01,
                    2.91225541e01,
                    2.91225457e01,
                    2.91225448e01,
                    2.91225407e01,
                    2.91225398e01,
                    2.91225393e01,
                    2.91225391e01,
                    2.91225390e01,
                    2.91225390e01,
                    2.91225389e01,
                    2.91225389e01,
                ],
            ],
        ),
        np.array(
            [
                [
                    7.82198834e-01,
                    2.84434275e-02,
                    3.42472318e-03,
                    1.45801296e-03,
                    2.02213149e-03,
                    2.00826891e-03,
                    1.27577246e-03,
                    4.27099544e-04,
                    2.07304219e-05,
                    4.88683542e-05,
                    4.14000873e-05,
                    1.57801531e-05,
                    2.08705254e-06,
                    2.27936213e-06,
                    1.09950377e-06,
                    7.09021331e-07,
                    7.09539552e-07,
                    7.08228574e-07,
                    7.08101644e-07,
                    7.08098642e-07,
                    7.08098641e-07,
                ],
                [
                    7.82198834e-01,
                    2.84434275e-02,
                    3.42472318e-03,
                    1.45801296e-03,
                    2.02213149e-03,
                    2.00826891e-03,
                    1.27577246e-03,
                    4.27099544e-04,
                    2.07304219e-05,
                    4.88683542e-05,
                    4.14000873e-05,
                    1.57801531e-05,
                    2.08705254e-06,
                    2.27936213e-06,
                    1.09950377e-06,
                    7.09021331e-07,
                    7.09539552e-07,
                    7.08228574e-07,
                    7.08101644e-07,
                    7.08098642e-07,
                    7.08098641e-07,
                ],
                [
                    8.94587126e-01,
                    4.07514176e-01,
                    3.66355642e-02,
                    1.22207690e-02,
                    5.91350591e-03,
                    3.17495745e-03,
                    1.50593011e-03,
                    5.80889579e-04,
                    2.44654124e-04,
                    1.40925782e-04,
                    6.26965047e-05,
                    1.91064075e-05,
                    9.45189755e-06,
                    6.16421495e-06,
                    5.17046678e-06,
                    4.24101973e-06,
                    3.82067554e-06,
                    3.55352666e-06,
                    3.38981154e-06,
                    3.29631224e-06,
                    3.25061722e-06,
                ],
                [
                    1.00000560e00,
                    7.63076662e-01,
                    6.77868181e-02,
                    2.23160672e-02,
                    9.56355479e-03,
                    4.26929321e-03,
                    1.72181442e-03,
                    7.25142247e-04,
                    4.54691084e-04,
                    2.27274223e-04,
                    8.54009496e-05,
                    2.26525796e-05,
                    1.63600080e-05,
                    9.80814666e-06,
                    8.98896697e-06,
                    7.55397947e-06,
                    6.73887287e-06,
                    6.22237216e-06,
                    5.90521384e-06,
                    5.72401646e-06,
                    5.63546023e-06,
                ],
                [
                    1.00000560e00,
                    7.63076662e-01,
                    6.77868181e-02,
                    2.23160672e-02,
                    9.56355479e-03,
                    4.26929321e-03,
                    1.72181442e-03,
                    7.25142247e-04,
                    4.54691084e-04,
                    2.27274223e-04,
                    8.54009496e-05,
                    2.26525796e-05,
                    1.63600080e-05,
                    9.80814666e-06,
                    8.98896697e-06,
                    7.55397947e-06,
                    6.73887287e-06,
                    6.22237216e-06,
                    5.90521384e-06,
                    5.72401646e-06,
                    5.63546023e-06,
                ],
            ],
        ),
    ]
    bayes.predictionIntervals.sld = [
        [
            np.array(
                [
                    [
                        -2.26996619e-06,
                        -2.26166520e-06,
                        -2.24783719e-06,
                        -2.22613645e-06,
                        -2.19406176e-06,
                        -2.17142372e-06,
                        1.88515086e-07,
                        2.60653226e-07,
                        3.44976651e-07,
                        4.40965268e-07,
                        5.53846231e-07,
                        6.92263150e-07,
                        8.41189317e-07,
                        9.28611746e-07,
                        8.59850711e-07,
                        6.40672491e-07,
                        4.45129915e-07,
                        4.69261576e-07,
                        7.08570915e-07,
                        9.66403060e-07,
                        1.08031232e-06,
                        1.05528544e-06,
                        9.86455522e-07,
                        9.38980618e-07,
                        9.20228966e-07,
                        9.15507876e-07,
                        9.14706158e-07,
                        9.14597081e-07,
                        9.14576711e-07,
                        9.14569865e-07,
                    ],
                    [
                        -2.26996619e-06,
                        -2.26166520e-06,
                        -2.24783719e-06,
                        -2.22613645e-06,
                        -2.19406176e-06,
                        -2.17142372e-06,
                        1.88515086e-07,
                        2.60653226e-07,
                        3.44976651e-07,
                        4.40965268e-07,
                        5.53846231e-07,
                        6.92263150e-07,
                        8.41189317e-07,
                        9.28611746e-07,
                        8.59850711e-07,
                        6.40672491e-07,
                        4.45129915e-07,
                        4.69261576e-07,
                        7.08570915e-07,
                        9.66403060e-07,
                        1.08031232e-06,
                        1.05528544e-06,
                        9.86455522e-07,
                        9.38980618e-07,
                        9.20228966e-07,
                        9.15507876e-07,
                        9.14706158e-07,
                        9.14597081e-07,
                        9.14576711e-07,
                        9.14569865e-07,
                    ],
                    [
                        -1.50598689e-07,
                        -1.46347158e-07,
                        -1.38880755e-07,
                        -1.07564801e-07,
                        1.23182525e-07,
                        6.23089296e-07,
                        2.07733185e-06,
                        2.14065378e-06,
                        2.16728982e-06,
                        2.04144854e-06,
                        1.74466846e-06,
                        1.85390758e-06,
                        2.36094420e-06,
                        2.55768452e-06,
                        2.25381054e-06,
                        1.86530516e-06,
                        1.68664783e-06,
                        1.84743812e-06,
                        2.30954159e-06,
                        2.85224030e-06,
                        3.22454811e-06,
                        3.35997774e-06,
                        3.36663861e-06,
                        3.34928647e-06,
                        3.34035624e-06,
                        3.33797637e-06,
                        3.33756709e-06,
                        3.33751126e-06,
                        3.33750083e-06,
                        3.33749733e-06,
                    ],
                    [
                        2.07300000e-06,
                        2.07300292e-06,
                        2.07379485e-06,
                        2.11519890e-06,
                        2.55438964e-06,
                        3.55503738e-06,
                        4.05904125e-06,
                        4.11311338e-06,
                        4.07922495e-06,
                        3.72064411e-06,
                        2.99405572e-06,
                        3.07268206e-06,
                        3.95544112e-06,
                        4.26687563e-06,
                        3.71632577e-06,
                        3.15016567e-06,
                        2.98922400e-06,
                        3.29339384e-06,
                        3.98924852e-06,
                        4.83082363e-06,
                        5.47423812e-06,
                        5.77801555e-06,
                        5.86387989e-06,
                        5.87813196e-06,
                        5.87950617e-06,
                        5.87958266e-06,
                        5.87958511e-06,
                        5.87958515e-06,
                        5.87958515e-06,
                        5.87958515e-06,
                    ],
                    [
                        2.07300000e-06,
                        2.07300292e-06,
                        2.07379485e-06,
                        2.11519890e-06,
                        2.55438964e-06,
                        3.55503738e-06,
                        4.05904125e-06,
                        4.11311338e-06,
                        4.07922495e-06,
                        3.72064411e-06,
                        2.99405572e-06,
                        3.07268206e-06,
                        3.95544112e-06,
                        4.26687563e-06,
                        3.71632577e-06,
                        3.15016567e-06,
                        2.98922400e-06,
                        3.29339384e-06,
                        3.98924852e-06,
                        4.83082363e-06,
                        5.47423812e-06,
                        5.77801555e-06,
                        5.86387989e-06,
                        5.87813196e-06,
                        5.87950617e-06,
                        5.87958266e-06,
                        5.87958511e-06,
                        5.87958515e-06,
                        5.87958515e-06,
                        5.87958515e-06,
                    ],
                ],
            ),
            np.array(
                [
                    [
                        -1.35107208e-06,
                        -1.33036284e-06,
                        -1.29371078e-06,
                        -1.23241324e-06,
                        -1.13554244e-06,
                        -3.51002390e-06,
                        5.85600156e-07,
                        8.57236782e-07,
                        1.19525576e-06,
                        1.57378055e-06,
                        1.91426623e-06,
                        2.08784412e-06,
                        2.00936631e-06,
                        1.71439287e-06,
                        1.27066669e-06,
                        7.62973092e-07,
                        4.60666931e-07,
                        6.93493469e-07,
                        1.45652624e-06,
                        2.42477950e-06,
                        3.34626572e-06,
                        4.13683755e-06,
                        4.64240264e-06,
                        4.65273375e-06,
                        4.65372768e-06,
                        4.65378295e-06,
                        4.65378471e-06,
                        4.65378475e-06,
                        4.65378475e-06,
                        4.65378475e-06,
                    ],
                    [
                        -1.35107208e-06,
                        -1.33036284e-06,
                        -1.29371078e-06,
                        -1.23241324e-06,
                        -1.13554244e-06,
                        -3.51002390e-06,
                        5.85600156e-07,
                        8.57236782e-07,
                        1.19525576e-06,
                        1.57378055e-06,
                        1.91426623e-06,
                        2.08784412e-06,
                        2.00936631e-06,
                        1.71439287e-06,
                        1.27066669e-06,
                        7.62973092e-07,
                        4.60666931e-07,
                        6.93493469e-07,
                        1.45652624e-06,
                        2.42477950e-06,
                        3.34626572e-06,
                        4.13683755e-06,
                        4.64240264e-06,
                        4.65273375e-06,
                        4.65372768e-06,
                        4.65378295e-06,
                        4.65378471e-06,
                        4.65378475e-06,
                        4.65378475e-06,
                        4.65378475e-06,
                    ],
                    [
                        3.19875093e-07,
                        3.30479403e-07,
                        3.49564160e-07,
                        3.97621442e-07,
                        6.24076436e-07,
                        -1.88708544e-07,
                        2.11120725e-06,
                        2.27195959e-06,
                        2.42630375e-06,
                        2.42402196e-06,
                        2.19947507e-06,
                        2.28976268e-06,
                        2.60261209e-06,
                        2.56234446e-06,
                        2.11162563e-06,
                        1.61774868e-06,
                        1.39431387e-06,
                        1.64217088e-06,
                        2.31360784e-06,
                        3.13090619e-06,
                        3.83908225e-06,
                        4.35302731e-06,
                        4.67870504e-06,
                        4.84354637e-06,
                        4.90376869e-06,
                        4.91908875e-06,
                        4.92196879e-06,
                        4.92248720e-06,
                        4.92261931e-06,
                        4.92266441e-06,
                    ],
                    [
                        2.07300000e-06,
                        2.07300241e-06,
                        2.07365590e-06,
                        2.10782176e-06,
                        2.47023394e-06,
                        3.29595019e-06,
                        3.71184421e-06,
                        3.75625892e-06,
                        3.71789509e-06,
                        3.31607852e-06,
                        2.49871058e-06,
                        2.50161165e-06,
                        3.22503390e-06,
                        3.45199857e-06,
                        2.99394321e-06,
                        2.51456240e-06,
                        2.37387787e-06,
                        2.63750456e-06,
                        3.21284100e-06,
                        3.87176041e-06,
                        4.35613565e-06,
                        4.57984935e-06,
                        4.71330577e-06,
                        5.02541464e-06,
                        5.14208904e-06,
                        5.17195834e-06,
                        5.17758173e-06,
                        5.17859423e-06,
                        5.17885226e-06,
                        5.17894035e-06,
                    ],
                    [
                        2.07300000e-06,
                        2.07300241e-06,
                        2.07365590e-06,
                        2.10782176e-06,
                        2.47023394e-06,
                        3.29595019e-06,
                        3.71184421e-06,
                        3.75625892e-06,
                        3.71789509e-06,
                        3.31607852e-06,
                        2.49871058e-06,
                        2.50161165e-06,
                        3.22503390e-06,
                        3.45199857e-06,
                        2.99394321e-06,
                        2.51456240e-06,
                        2.37387787e-06,
                        2.63750456e-06,
                        3.21284100e-06,
                        3.87176041e-06,
                        4.35613565e-06,
                        4.57984935e-06,
                        4.71330577e-06,
                        5.02541464e-06,
                        5.14208904e-06,
                        5.17195834e-06,
                        5.17758173e-06,
                        5.17859423e-06,
                        5.17885226e-06,
                        5.17894035e-06,
                    ],
                ],
            ),
        ],
    ]
    bayes.predictionIntervals.reflectivityXData = [
        np.array(
            [
                [
                    0.011403,
                    0.013861,
                    0.016848,
                    0.020479,
                    0.024892,
                    0.030256,
                    0.036777,
                    0.044702,
                    0.054336,
                    0.066045,
                    0.080279,
                    0.097579,
                    0.11861,
                    0.14417,
                    0.17524,
                    0.213,
                    0.25891,
                    0.3147,
                    0.38252,
                    0.46496,
                    0.56516,
                ],
            ],
        ),
        np.array(
            [
                [
                    0.011403,
                    0.013861,
                    0.016848,
                    0.020479,
                    0.024892,
                    0.030256,
                    0.036777,
                    0.044702,
                    0.054336,
                    0.066045,
                    0.080279,
                    0.097579,
                    0.11861,
                    0.14417,
                    0.17524,
                    0.213,
                    0.25891,
                    0.3147,
                    0.38252,
                    0.46496,
                    0.56516,
                ],
            ],
        ),
    ]
    bayes.predictionIntervals.sldXData = [
        [
            np.array(
                [
                    [
                        0.0,
                        11.0,
                        22.0,
                        33.0,
                        44.0,
                        55.0,
                        66.0,
                        77.0,
                        88.0,
                        99.0,
                        110.0,
                        121.0,
                        132.0,
                        143.0,
                        154.0,
                        165.0,
                        176.0,
                        187.0,
                        198.0,
                        209.0,
                        220.0,
                        231.0,
                        242.0,
                        253.0,
                        264.0,
                        275.0,
                        286.0,
                        297.0,
                        308.0,
                        319.0,
                    ],
                ],
            ),
            np.array(
                [
                    [
                        0.0,
                        11.0,
                        22.0,
                        33.0,
                        44.0,
                        55.0,
                        66.0,
                        77.0,
                        88.0,
                        99.0,
                        110.0,
                        121.0,
                        132.0,
                        143.0,
                        154.0,
                        165.0,
                        176.0,
                        187.0,
                        198.0,
                        209.0,
                        220.0,
                        231.0,
                        242.0,
                        253.0,
                        264.0,
                        275.0,
                        286.0,
                        297.0,
                        308.0,
                        319.0,
                    ],
                ],
            ),
        ],
    ]
    bayes.predictionIntervals.sampleChi = np.array(
        [
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.46133559e16,
            1.12748255e06,
            1.12748255e06,
            1.46133559e16,
        ],
    )
    bayes.confidenceIntervals = RATapi.rat_core.ConfidenceIntervals()
    bayes.confidenceIntervals.percentile65 = np.array(
        [
            [
                -1.29074586e-231,
                8.33251318e000,
                1.75397363e001,
                1.75397363e001,
                9.85302945e000,
                9.85302945e000,
                8.34197863e000,
                8.34197863e000,
                1.65750684e001,
                1.45435510e001,
                1.45435510e001,
                1.52609047e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                7.08098641e-007,
                7.08098641e-007,
                4.65378475e-006,
            ],
            [
                8.33251318e000,
                5.48185565e001,
                5.48185565e001,
                4.57554170e001,
                4.57554170e001,
                1.17557273e001,
                1.17557273e001,
                3.18752608e001,
                3.18752608e001,
                1.65750684e001,
                1.52609047e001,
                4.88237113e001,
                4.88237113e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                5.87958515e-006,
                5.87958515e-006,
            ],
        ],
    )
    bayes.confidenceIntervals.percentile95 = np.array(
        [
            [
                -1.29074586e-231,
                8.33251318e000,
                1.75397363e001,
                1.75397363e001,
                9.85302945e000,
                9.85302945e000,
                8.34197863e000,
                8.34197863e000,
                1.65750684e001,
                1.45435510e001,
                1.45435510e001,
                1.52609047e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                7.08098641e-007,
                7.08098641e-007,
                4.65378475e-006,
            ],
            [
                8.33251318e000,
                5.48185565e001,
                5.48185565e001,
                4.57554170e001,
                4.57554170e001,
                1.17557273e001,
                1.17557273e001,
                3.18752608e001,
                3.18752608e001,
                1.65750684e001,
                1.52609047e001,
                4.88237113e001,
                4.88237113e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                5.87958515e-006,
                5.87958515e-006,
            ],
        ],
    )
    bayes.confidenceIntervals.mean = np.array(
        [
            [
                4.16625659e00,
                3.15755349e01,
                3.61791464e01,
                3.16475766e01,
                2.78042232e01,
                1.08043784e01,
                1.00488530e01,
                2.01086197e01,
                2.42251646e01,
                1.55593097e01,
                1.49022278e01,
                3.20423080e01,
                4.85551946e01,
                3.87046084e01,
                1.45612723e01,
                3.15508089e-06,
                3.29384190e-06,
                5.26668495e-06,
            ],
        ],
    )
    bayes.dreamParams = RATapi.rat_core.DreamParams()
    bayes.dreamParams.nParams = 18.0
    bayes.dreamParams.nChains = 1.0
    bayes.dreamParams.nGenerations = 1.0
    bayes.dreamParams.parallel = 0
    bayes.dreamParams.CPU = 1.0
    bayes.dreamParams.jumpProbability = 0.5
    bayes.dreamParams.pUnitGamma = 0.2
    bayes.dreamParams.nCR = 3.0
    bayes.dreamParams.delta = 3.0
    bayes.dreamParams.steps = 50.0
    bayes.dreamParams.zeta = 1e-12
    bayes.dreamParams.outlier = "iqr"
    bayes.dreamParams.adaptPCR = 0
    bayes.dreamParams.thinning = 1.0
    bayes.dreamParams.epsilon = 0.025
    bayes.dreamParams.ABC = 0
    bayes.dreamParams.IO = 0
    bayes.dreamParams.storeOutput = 0
    bayes.dreamParams.R = np.array([[0.0]])
    bayes.dreamOutput = RATapi.rat_core.DreamOutput()
    bayes.dreamOutput.allChains = np.array(
        [
            [
                [8.33251318e00],
                [5.48185565e01],
                [1.75397363e01],
                [4.57554170e01],
                [9.85302945e00],
                [1.17557273e01],
                [8.34197863e00],
                [3.18752608e01],
                [1.65750684e01],
                [1.45435510e01],
                [1.52609047e01],
                [4.88237113e01],
                [4.82866779e01],
                [2.91225389e01],
                [5.60206314e-06],
                [7.08098641e-07],
                [5.87958515e-06],
                [4.65378475e-06],
                [-4.86509830e01],
                [-5.63741277e05],
            ],
        ],
    )
    bayes.dreamOutput.outlierChains = np.array([[0.0, 0.0]])
    bayes.dreamOutput.runtime = 2.6e-06
    bayes.dreamOutput.iteration = 2.0
    bayes.dreamOutput.modelOutput = 0.0
    bayes.dreamOutput.AR = np.array([[1.0, np.nan]])
    bayes.dreamOutput.R_stat = np.array(
        [
            [
                1.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ],
        ],
    )
    bayes.dreamOutput.CR = np.array([[1.00000000, 0.33333333, 0.33333333, 0.33333333]])
    bayes.nestedSamplerOutput = RATapi.rat_core.NestedSamplerOutput()
    bayes.nestedSamplerOutput.logZ = 0.0
    bayes.nestedSamplerOutput.nestSamples = np.array([[0.0, 0.0]])
    bayes.nestedSamplerOutput.postSamples = np.array([[0.0, 0.0]])
    bayes.chain = np.array(
        [
            [
                -1.29231905e-231,
                8.33251318e000,
                5.48185565e001,
                1.75397363e001,
                4.57554170e001,
                9.85302945e000,
                1.17557273e001,
                8.34197863e000,
                3.18752608e001,
                1.65750684e001,
                1.45435510e001,
                1.52609047e001,
                4.88237113e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                7.08098641e-007,
                5.87958515e-006,
            ],
            [
                8.33251318e000,
                5.48185565e001,
                1.75397363e001,
                4.57554170e001,
                9.85302945e000,
                1.17557273e001,
                8.34197863e000,
                3.18752608e001,
                1.65750684e001,
                1.45435510e001,
                1.52609047e001,
                4.88237113e001,
                4.82866779e001,
                2.91225389e001,
                5.60206314e-006,
                7.08098641e-007,
                5.87958515e-006,
                4.65378475e-006,
            ],
        ],
    )

    return bayes


@pytest.fixture
def dream_results():
    """The python BayesResults object for a Dream optimisation of the project set out in "DSPC_standard_layers.py".

    This optimisation used the parameters: nSamples=1, nChains=1. However, the calculationResults, contrastParams,
    and fitParams are taken from an optimisation with the parameters: nSamples=50000, nChains=10.
    """
    return RATapi.outputs.BayesResults(
        reflectivity=[
            np.array(
                [
                    [1.14030000e-02, 1.45891447e01],
                    [1.38610000e-02, 1.45739839e01],
                    [1.68480000e-02, 1.45684725e01],
                    [2.04790000e-02, 1.45658452e01],
                    [2.48920000e-02, 1.45641513e01],
                    [3.02560000e-02, 1.45628311e01],
                    [3.67770000e-02, 1.45618510e01],
                    [4.47020000e-02, 1.45613427e01],
                    [5.43360000e-02, 1.45612940e01],
                    [6.60450000e-02, 1.45613971e01],
                    [8.02790000e-02, 1.45613678e01],
                    [9.75790000e-02, 1.45612960e01],
                    [1.18610000e-01, 1.45612799e01],
                    [1.44170000e-01, 1.45612749e01],
                    [1.75240000e-01, 1.45612726e01],
                    [2.13000000e-01, 1.45612723e01],
                    [2.58910000e-01, 1.45612724e01],
                    [3.14700000e-01, 1.45612723e01],
                    [3.82520000e-01, 1.45612723e01],
                    [4.64960000e-01, 1.45612723e01],
                    [5.65160000e-01, 1.45612723e01],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.00000316e00],
                    [1.38610000e-02, 1.21238320e-01],
                    [1.68480000e-02, 2.13433068e-02],
                    [2.04790000e-02, 8.71162144e-03],
                    [2.48920000e-02, 5.72807902e-03],
                    [3.02560000e-02, 3.71827525e-03],
                    [3.67770000e-02, 1.72067772e-03],
                    [4.47020000e-02, 3.51471141e-04],
                    [5.43360000e-02, 2.03975460e-05],
                    [6.60450000e-02, 1.89221181e-04],
                    [8.02790000e-02, 1.49280488e-04],
                    [9.75790000e-02, 3.86439083e-05],
                    [1.18610000e-01, 1.63612094e-05],
                    [1.44170000e-01, 6.75770325e-06],
                    [1.75240000e-01, 3.45955607e-06],
                    [2.13000000e-01, 3.20880782e-06],
                    [2.58910000e-01, 3.32068345e-06],
                    [3.14700000e-01, 3.21671807e-06],
                    [3.82520000e-01, 3.15612021e-06],
                    [4.64960000e-01, 3.15645546e-06],
                    [5.65160000e-01, 3.15511510e-06],
                ],
            ),
        ],
        simulation=[
            np.array(
                [
                    [1.14030000e-02, 1.45891447e01],
                    [1.38610000e-02, 1.45739839e01],
                    [1.68480000e-02, 1.45684725e01],
                    [2.04790000e-02, 1.45658452e01],
                    [2.48920000e-02, 1.45641513e01],
                    [3.02560000e-02, 1.45628311e01],
                    [3.67770000e-02, 1.45618510e01],
                    [4.47020000e-02, 1.45613427e01],
                    [5.43360000e-02, 1.45612940e01],
                    [6.60450000e-02, 1.45613971e01],
                    [8.02790000e-02, 1.45613678e01],
                    [9.75790000e-02, 1.45612960e01],
                    [1.18610000e-01, 1.45612799e01],
                    [1.44170000e-01, 1.45612749e01],
                    [1.75240000e-01, 1.45612726e01],
                    [2.13000000e-01, 1.45612723e01],
                    [2.58910000e-01, 1.45612724e01],
                    [3.14700000e-01, 1.45612723e01],
                    [3.82520000e-01, 1.45612723e01],
                    [4.64960000e-01, 1.45612723e01],
                    [5.65160000e-01, 1.45612723e01],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.00000316e00],
                    [1.38610000e-02, 1.21238320e-01],
                    [1.68480000e-02, 2.13433068e-02],
                    [2.04790000e-02, 8.71162144e-03],
                    [2.48920000e-02, 5.72807902e-03],
                    [3.02560000e-02, 3.71827525e-03],
                    [3.67770000e-02, 1.72067772e-03],
                    [4.47020000e-02, 3.51471141e-04],
                    [5.43360000e-02, 2.03975460e-05],
                    [6.60450000e-02, 1.89221181e-04],
                    [8.02790000e-02, 1.49280488e-04],
                    [9.75790000e-02, 3.86439083e-05],
                    [1.18610000e-01, 1.63612094e-05],
                    [1.44170000e-01, 6.75770325e-06],
                    [1.75240000e-01, 3.45955607e-06],
                    [2.13000000e-01, 3.20880782e-06],
                    [2.58910000e-01, 3.32068345e-06],
                    [3.14700000e-01, 3.21671807e-06],
                    [3.82520000e-01, 3.15612021e-06],
                    [4.64960000e-01, 3.15645546e-06],
                    [5.65160000e-01, 3.15511510e-06],
                ],
            ),
        ],
        shiftedData=[
            np.array(
                [
                    [1.1403e-02, 1.0063e00, 1.9003e-02],
                    [1.3861e-02, 9.0118e-01, 1.1774e-02],
                    [1.6848e-02, 7.0455e-02, 1.3083e-03],
                    [2.0479e-02, 1.7544e-02, 5.1254e-04],
                    [2.4892e-02, 6.4257e-03, 2.6236e-04],
                    [3.0256e-02, 2.7746e-03, 8.5758e-05],
                    [3.6777e-02, 1.8591e-03, 4.9391e-05],
                    [4.4702e-02, 1.1002e-03, 3.2644e-05],
                    [5.4336e-02, 6.6691e-04, 2.1365e-05],
                    [6.6045e-02, 6.0729e-04, 1.1791e-05],
                    [8.0279e-02, 5.8755e-04, 1.5569e-05],
                    [9.7579e-02, 3.2700e-04, 6.5280e-06],
                    [1.1861e-01, 7.8205e-05, 2.5881e-06],
                    [1.4417e-01, 3.3455e-05, 1.5143e-06],
                    [1.7524e-01, 4.9313e-06, 5.6663e-07],
                    [2.1300e-01, 4.1948e-06, 6.8549e-07],
                    [2.5891e-01, 3.9863e-06, 8.2061e-07],
                    [3.1470e-01, 2.0861e-06, 6.1379e-07],
                    [3.8252e-01, 2.1154e-06, 6.3084e-07],
                    [4.6496e-01, 1.9906e-06, 6.0793e-07],
                    [5.6516e-01, 2.3816e-06, 7.0610e-07],
                ],
            ),
            np.array(
                [
                    [1.14030000e-02, 1.20493333e-02, 7.32200000e-04],
                    [1.38610000e-02, 6.71800000e-03, 3.71853333e-04],
                    [1.68480000e-02, 4.10506667e-03, 2.23106667e-04],
                    [2.04790000e-02, 2.43306667e-03, 1.46873333e-04],
                    [2.48920000e-02, 1.40940000e-03, 9.59200000e-05],
                    [3.02560000e-02, 7.43400000e-04, 3.50226667e-05],
                    [3.67770000e-02, 3.51466667e-04, 1.59573333e-05],
                    [4.47020000e-02, 9.80666667e-05, 7.29466667e-06],
                    [5.43360000e-02, 1.73500000e-05, 2.59200000e-06],
                    [6.60450000e-02, 4.86286667e-05, 2.46453333e-06],
                    [8.02790000e-02, 9.71733333e-05, 4.68166667e-06],
                    [9.75790000e-02, 7.06533333e-05, 2.32180000e-06],
                    [1.18610000e-01, 1.93046667e-05, 1.00813333e-06],
                    [1.44170000e-01, 6.61193333e-06, 5.29773333e-07],
                    [1.75240000e-01, 3.23726667e-06, 3.65933333e-07],
                    [2.13000000e-01, 3.29920000e-06, 4.74273333e-07],
                    [2.58910000e-01, 1.71180000e-06, 4.24720000e-07],
                    [3.14700000e-01, 2.24020000e-06, 5.08946667e-07],
                    [3.82520000e-01, 2.73306667e-06, 5.71513333e-07],
                    [4.64960000e-01, 2.36153333e-06, 5.24806667e-07],
                    [5.65160000e-01, 2.17460000e-06, 5.31346667e-07],
                ],
            ),
        ],
        layerSlds=[
            [
                np.array(
                    [
                        [3.15755349e01, 3.35278238e-06, 4.16625659e00],
                        [3.61791464e01, 7.68327921e-07, 4.16625659e00],
                        [1.00488530e01, 2.06044530e-06, 2.78042232e01],
                        [1.08043784e01, 3.29384190e-06, 2.78042232e01],
                        [2.42251646e01, 2.35556998e-06, 1.55593097e01],
                        [1.49022278e01, 7.42138004e-07, 1.55593097e01],
                        [1.49022278e01, 7.42138004e-07, 1.55593097e01],
                        [2.42251646e01, 2.35556998e-06, 1.55593097e01],
                    ],
                ),
            ],
            [
                np.array(
                    [
                        [3.15755349e01, 4.11636356e-06, 4.16625659e00],
                        [3.61791464e01, 1.39268494e-06, 4.16625659e00],
                        [1.00488530e01, 2.45715680e-06, 2.78042232e01],
                        [1.08043784e01, 5.26668495e-06, 2.78042232e01],
                        [2.42251646e01, 3.31348777e-06, 1.55593097e01],
                        [1.49022278e01, 1.37428245e-06, 1.55593097e01],
                        [1.49022278e01, 1.37428245e-06, 1.55593097e01],
                        [2.42251646e01, 3.31348777e-06, 1.55593097e01],
                    ],
                ),
            ],
        ],
        sldProfiles=[
            [
                np.array(
                    [
                        [0.00000000e00, 2.07301741e-06],
                        [1.10000000e01, 2.07309604e-06],
                        [2.20000000e01, 2.07345780e-06],
                        [3.30000000e01, 2.07491687e-06],
                        [4.40000000e01, 2.17562259e-06],
                        [5.50000000e01, 3.22650495e-06],
                        [6.60000000e01, 3.40913848e-06],
                        [7.70000000e01, 3.13506253e-06],
                        [8.80000000e01, 1.20456289e-06],
                        [9.90000000e01, 1.27138002e-06],
                        [1.10000000e02, 1.56285495e-06],
                        [1.21000000e02, 1.84518709e-06],
                        [1.32000000e02, 2.00498449e-06],
                        [1.43000000e02, 1.96260665e-06],
                        [1.54000000e02, 1.71966879e-06],
                        [1.65000000e02, 1.39383267e-06],
                        [1.76000000e02, 1.22970195e-06],
                        [1.87000000e02, 1.41565526e-06],
                        [1.98000000e02, 1.88172657e-06],
                        [2.09000000e02, 2.40655867e-06],
                        [2.20000000e02, 2.83570420e-06],
                        [2.31000000e02, 3.11210082e-06],
                        [2.42000000e02, 3.24277802e-06],
                        [2.53000000e02, 3.28427452e-06],
                        [2.64000000e02, 3.29268847e-06],
                        [2.75000000e02, 3.29375418e-06],
                        [2.86000000e02, 3.29383773e-06],
                        [2.97000000e02, 3.29384178e-06],
                        [3.08000000e02, 3.29384190e-06],
                    ],
                ),
                np.array(
                    [
                        [0.00000000e00, 2.07301819e-06],
                        [1.10000000e01, 2.07310296e-06],
                        [2.20000000e01, 2.07350415e-06],
                        [3.30000000e01, 2.07518460e-06],
                        [4.40000000e01, 2.23394767e-06],
                        [5.50000000e01, 3.90646736e-06],
                        [6.60000000e01, 4.18619279e-06],
                        [7.70000000e01, 3.91675004e-06],
                        [8.80000000e01, 1.92451547e-06],
                        [9.90000000e01, 2.06988861e-06],
                        [1.10000000e02, 2.47614934e-06],
                        [1.21000000e02, 2.84889288e-06],
                        [1.32000000e02, 3.00529286e-06],
                        [1.43000000e02, 2.86310144e-06],
                        [1.54000000e02, 2.49967205e-06],
                        [1.65000000e02, 2.09673058e-06],
                        [1.76000000e02, 1.92431424e-06],
                        [1.87000000e02, 2.18829605e-06],
                        [1.98000000e02, 2.83032190e-06],
                        [2.09000000e02, 3.62584708e-06],
                        [2.20000000e02, 4.36871153e-06],
                        [2.31000000e02, 4.89806866e-06],
                        [2.42000000e02, 5.16145928e-06],
                        [2.53000000e02, 5.24684298e-06],
                        [2.64000000e02, 5.26428707e-06],
                        [2.75000000e02, 5.26650242e-06],
                        [2.86000000e02, 5.26667628e-06],
                        [2.97000000e02, 5.26668469e-06],
                        [3.08000000e02, 5.26668495e-06],
                    ],
                ),
            ],
        ],
        resampledLayers=[[np.array([[0.0, 0.0, 0.0]])], [np.array([[0.0, 0.0, 0.0]])]],
        calculationResults=RATapi.outputs.CalculationResults(
            chiValues=np.array([4.6077885, 7.00028098]),
            sumChi=11.608069475997699,
        ),
        contrastParams=RATapi.outputs.ContrastParams(
            backgroundParams=np.array([2.37113128e-06, 1.99006694e-06]),
            scalefactors=np.array([0.1, 0.15]),
            bulkIn=np.array([2.073e-06, 2.073e-06]),
            bulkOut=np.array([6.01489149e-06, 1.59371685e-06]),
            resolutionParams=np.array([0.03, 0.03]),
            subRoughs=np.array([6.19503045, 6.19503045]),
            resample=np.array([0.0, 0.0]),
        ),
        fitParams=np.array(
            [
                6.19503045e00,
                1.84420960e01,
                2.11039621e01,
                8.75538121e00,
                3.72292994e00,
                1.84624551e01,
                1.02316734e01,
                2.31156093e01,
                1.09906265e01,
                5.71005361e00,
                1.67933822e01,
                1.72009856e01,
                3.00260126e01,
                2.94448999e01,
                2.37113128e-06,
                1.99006694e-06,
                6.01489149e-06,
                1.59371685e-06,
            ],
        ),
        fitNames=[
            "Substrate Roughness",
            "Oxide Thickness",
            "SAM Tails Thickness",
            "SAM Tails Hydration",
            "SAM Roughness",
            "CW Thickness",
            "SAM Heads Thickness",
            "SAM Heads Hydration",
            "Bilayer Heads Thickness",
            "Bilayer Roughness",
            "Bilayer Tails Thickness",
            "Bilayer Tails Hydration",
            "Bilayer Heads Hydration",
            "Oxide Hydration",
            "Background parameter D2O",
            "Background parameter SMW",
            "D2O",
            "SMW",
        ],
        predictionIntervals=RATapi.outputs.PredictionIntervals(
            reflectivity=[
                np.array(
                    [
                        [
                            1.00000560e00,
                            7.07687612e-01,
                            2.08315160e-02,
                            2.40787966e-03,
                            2.17627660e-03,
                            2.54301700e-03,
                            1.76309827e-03,
                            6.15269679e-04,
                            3.43679710e-05,
                            7.93625275e-05,
                            1.27760549e-04,
                            3.35799941e-05,
                            6.75048751e-06,
                            7.64258431e-06,
                            6.05800395e-06,
                            5.60288298e-06,
                            5.60333677e-06,
                            5.60223341e-06,
                            5.60206667e-06,
                            5.60206314e-06,
                            5.60206314e-06,
                        ],
                        [
                            1.00000560e00,
                            7.07687612e-01,
                            2.08315160e-02,
                            2.40787966e-03,
                            2.17627660e-03,
                            2.54301700e-03,
                            1.76309827e-03,
                            6.15269679e-04,
                            3.43679710e-05,
                            7.93625275e-05,
                            1.27760549e-04,
                            3.35799941e-05,
                            6.75048751e-06,
                            7.64258431e-06,
                            6.05800395e-06,
                            5.60288298e-06,
                            5.60333677e-06,
                            5.60223341e-06,
                            5.60206667e-06,
                            5.60206314e-06,
                            5.60206314e-06,
                        ],
                        [
                            1.55201098e01,
                            1.53748894e01,
                            1.50402011e01,
                            1.50299478e01,
                            1.50290525e01,
                            1.50287935e01,
                            1.50282126e01,
                            1.50276044e01,
                            1.50273248e01,
                            1.50273185e01,
                            1.50272997e01,
                            1.50272498e01,
                            1.50272364e01,
                            1.50272347e01,
                            1.50272335e01,
                            1.50272330e01,
                            1.50272329e01,
                            1.50272328e01,
                            1.50272328e01,
                            1.50272328e01,
                            1.50272328e01,
                        ],
                        [
                            2.91397425e01,
                            2.91324974e01,
                            2.91281368e01,
                            2.91255472e01,
                            2.91240295e01,
                            2.91231836e01,
                            2.91227893e01,
                            2.91226872e01,
                            2.91226902e01,
                            2.91226359e01,
                            2.91225541e01,
                            2.91225457e01,
                            2.91225448e01,
                            2.91225407e01,
                            2.91225398e01,
                            2.91225393e01,
                            2.91225391e01,
                            2.91225390e01,
                            2.91225390e01,
                            2.91225389e01,
                            2.91225389e01,
                        ],
                        [
                            2.91397425e01,
                            2.91324974e01,
                            2.91281368e01,
                            2.91255472e01,
                            2.91240295e01,
                            2.91231836e01,
                            2.91227893e01,
                            2.91226872e01,
                            2.91226902e01,
                            2.91226359e01,
                            2.91225541e01,
                            2.91225457e01,
                            2.91225448e01,
                            2.91225407e01,
                            2.91225398e01,
                            2.91225393e01,
                            2.91225391e01,
                            2.91225390e01,
                            2.91225390e01,
                            2.91225389e01,
                            2.91225389e01,
                        ],
                    ],
                ),
                np.array(
                    [
                        [
                            7.82198834e-01,
                            2.84434275e-02,
                            3.42472318e-03,
                            1.45801296e-03,
                            2.02213149e-03,
                            2.00826891e-03,
                            1.27577246e-03,
                            4.27099544e-04,
                            2.07304219e-05,
                            4.88683542e-05,
                            4.14000873e-05,
                            1.57801531e-05,
                            2.08705254e-06,
                            2.27936213e-06,
                            1.09950377e-06,
                            7.09021331e-07,
                            7.09539552e-07,
                            7.08228574e-07,
                            7.08101644e-07,
                            7.08098642e-07,
                            7.08098641e-07,
                        ],
                        [
                            7.82198834e-01,
                            2.84434275e-02,
                            3.42472318e-03,
                            1.45801296e-03,
                            2.02213149e-03,
                            2.00826891e-03,
                            1.27577246e-03,
                            4.27099544e-04,
                            2.07304219e-05,
                            4.88683542e-05,
                            4.14000873e-05,
                            1.57801531e-05,
                            2.08705254e-06,
                            2.27936213e-06,
                            1.09950377e-06,
                            7.09021331e-07,
                            7.09539552e-07,
                            7.08228574e-07,
                            7.08101644e-07,
                            7.08098642e-07,
                            7.08098641e-07,
                        ],
                        [
                            8.94587126e-01,
                            4.07514176e-01,
                            3.66355642e-02,
                            1.22207690e-02,
                            5.91350591e-03,
                            3.17495745e-03,
                            1.50593011e-03,
                            5.80889579e-04,
                            2.44654124e-04,
                            1.40925782e-04,
                            6.26965047e-05,
                            1.91064075e-05,
                            9.45189755e-06,
                            6.16421495e-06,
                            5.17046678e-06,
                            4.24101973e-06,
                            3.82067554e-06,
                            3.55352666e-06,
                            3.38981154e-06,
                            3.29631224e-06,
                            3.25061722e-06,
                        ],
                        [
                            1.00000560e00,
                            7.63076662e-01,
                            6.77868181e-02,
                            2.23160672e-02,
                            9.56355479e-03,
                            4.26929321e-03,
                            1.72181442e-03,
                            7.25142247e-04,
                            4.54691084e-04,
                            2.27274223e-04,
                            8.54009496e-05,
                            2.26525796e-05,
                            1.63600080e-05,
                            9.80814666e-06,
                            8.98896697e-06,
                            7.55397947e-06,
                            6.73887287e-06,
                            6.22237216e-06,
                            5.90521384e-06,
                            5.72401646e-06,
                            5.63546023e-06,
                        ],
                        [
                            1.00000560e00,
                            7.63076662e-01,
                            6.77868181e-02,
                            2.23160672e-02,
                            9.56355479e-03,
                            4.26929321e-03,
                            1.72181442e-03,
                            7.25142247e-04,
                            4.54691084e-04,
                            2.27274223e-04,
                            8.54009496e-05,
                            2.26525796e-05,
                            1.63600080e-05,
                            9.80814666e-06,
                            8.98896697e-06,
                            7.55397947e-06,
                            6.73887287e-06,
                            6.22237216e-06,
                            5.90521384e-06,
                            5.72401646e-06,
                            5.63546023e-06,
                        ],
                    ],
                ),
            ],
            sld=[
                [
                    np.array(
                        [
                            [
                                -2.26996619e-06,
                                -2.26166520e-06,
                                -2.24783719e-06,
                                -2.22613645e-06,
                                -2.19406176e-06,
                                -2.17142372e-06,
                                1.88515086e-07,
                                2.60653226e-07,
                                3.44976651e-07,
                                4.40965268e-07,
                                5.53846231e-07,
                                6.92263150e-07,
                                8.41189317e-07,
                                9.28611746e-07,
                                8.59850711e-07,
                                6.40672491e-07,
                                4.45129915e-07,
                                4.69261576e-07,
                                7.08570915e-07,
                                9.66403060e-07,
                                1.08031232e-06,
                                1.05528544e-06,
                                9.86455522e-07,
                                9.38980618e-07,
                                9.20228966e-07,
                                9.15507876e-07,
                                9.14706158e-07,
                                9.14597081e-07,
                                9.14576711e-07,
                                9.14569865e-07,
                            ],
                            [
                                -2.26996619e-06,
                                -2.26166520e-06,
                                -2.24783719e-06,
                                -2.22613645e-06,
                                -2.19406176e-06,
                                -2.17142372e-06,
                                1.88515086e-07,
                                2.60653226e-07,
                                3.44976651e-07,
                                4.40965268e-07,
                                5.53846231e-07,
                                6.92263150e-07,
                                8.41189317e-07,
                                9.28611746e-07,
                                8.59850711e-07,
                                6.40672491e-07,
                                4.45129915e-07,
                                4.69261576e-07,
                                7.08570915e-07,
                                9.66403060e-07,
                                1.08031232e-06,
                                1.05528544e-06,
                                9.86455522e-07,
                                9.38980618e-07,
                                9.20228966e-07,
                                9.15507876e-07,
                                9.14706158e-07,
                                9.14597081e-07,
                                9.14576711e-07,
                                9.14569865e-07,
                            ],
                            [
                                -1.50598689e-07,
                                -1.46347158e-07,
                                -1.38880755e-07,
                                -1.07564801e-07,
                                1.23182525e-07,
                                6.23089296e-07,
                                2.07733185e-06,
                                2.14065378e-06,
                                2.16728982e-06,
                                2.04144854e-06,
                                1.74466846e-06,
                                1.85390758e-06,
                                2.36094420e-06,
                                2.55768452e-06,
                                2.25381054e-06,
                                1.86530516e-06,
                                1.68664783e-06,
                                1.84743812e-06,
                                2.30954159e-06,
                                2.85224030e-06,
                                3.22454811e-06,
                                3.35997774e-06,
                                3.36663861e-06,
                                3.34928647e-06,
                                3.34035624e-06,
                                3.33797637e-06,
                                3.33756709e-06,
                                3.33751126e-06,
                                3.33750083e-06,
                                3.33749733e-06,
                            ],
                            [
                                2.07300000e-06,
                                2.07300292e-06,
                                2.07379485e-06,
                                2.11519890e-06,
                                2.55438964e-06,
                                3.55503738e-06,
                                4.05904125e-06,
                                4.11311338e-06,
                                4.07922495e-06,
                                3.72064411e-06,
                                2.99405572e-06,
                                3.07268206e-06,
                                3.95544112e-06,
                                4.26687563e-06,
                                3.71632577e-06,
                                3.15016567e-06,
                                2.98922400e-06,
                                3.29339384e-06,
                                3.98924852e-06,
                                4.83082363e-06,
                                5.47423812e-06,
                                5.77801555e-06,
                                5.86387989e-06,
                                5.87813196e-06,
                                5.87950617e-06,
                                5.87958266e-06,
                                5.87958511e-06,
                                5.87958515e-06,
                                5.87958515e-06,
                                5.87958515e-06,
                            ],
                            [
                                2.07300000e-06,
                                2.07300292e-06,
                                2.07379485e-06,
                                2.11519890e-06,
                                2.55438964e-06,
                                3.55503738e-06,
                                4.05904125e-06,
                                4.11311338e-06,
                                4.07922495e-06,
                                3.72064411e-06,
                                2.99405572e-06,
                                3.07268206e-06,
                                3.95544112e-06,
                                4.26687563e-06,
                                3.71632577e-06,
                                3.15016567e-06,
                                2.98922400e-06,
                                3.29339384e-06,
                                3.98924852e-06,
                                4.83082363e-06,
                                5.47423812e-06,
                                5.77801555e-06,
                                5.86387989e-06,
                                5.87813196e-06,
                                5.87950617e-06,
                                5.87958266e-06,
                                5.87958511e-06,
                                5.87958515e-06,
                                5.87958515e-06,
                                5.87958515e-06,
                            ],
                        ]
                    ),
                    np.array(
                        [
                            [
                                -1.35107208e-06,
                                -1.33036284e-06,
                                -1.29371078e-06,
                                -1.23241324e-06,
                                -1.13554244e-06,
                                -3.51002390e-06,
                                5.85600156e-07,
                                8.57236782e-07,
                                1.19525576e-06,
                                1.57378055e-06,
                                1.91426623e-06,
                                2.08784412e-06,
                                2.00936631e-06,
                                1.71439287e-06,
                                1.27066669e-06,
                                7.62973092e-07,
                                4.60666931e-07,
                                6.93493469e-07,
                                1.45652624e-06,
                                2.42477950e-06,
                                3.34626572e-06,
                                4.13683755e-06,
                                4.64240264e-06,
                                4.65273375e-06,
                                4.65372768e-06,
                                4.65378295e-06,
                                4.65378471e-06,
                                4.65378475e-06,
                                4.65378475e-06,
                                4.65378475e-06,
                            ],
                            [
                                -1.35107208e-06,
                                -1.33036284e-06,
                                -1.29371078e-06,
                                -1.23241324e-06,
                                -1.13554244e-06,
                                -3.51002390e-06,
                                5.85600156e-07,
                                8.57236782e-07,
                                1.19525576e-06,
                                1.57378055e-06,
                                1.91426623e-06,
                                2.08784412e-06,
                                2.00936631e-06,
                                1.71439287e-06,
                                1.27066669e-06,
                                7.62973092e-07,
                                4.60666931e-07,
                                6.93493469e-07,
                                1.45652624e-06,
                                2.42477950e-06,
                                3.34626572e-06,
                                4.13683755e-06,
                                4.64240264e-06,
                                4.65273375e-06,
                                4.65372768e-06,
                                4.65378295e-06,
                                4.65378471e-06,
                                4.65378475e-06,
                                4.65378475e-06,
                                4.65378475e-06,
                            ],
                            [
                                3.19875093e-07,
                                3.30479403e-07,
                                3.49564160e-07,
                                3.97621442e-07,
                                6.24076436e-07,
                                -1.88708544e-07,
                                2.11120725e-06,
                                2.27195959e-06,
                                2.42630375e-06,
                                2.42402196e-06,
                                2.19947507e-06,
                                2.28976268e-06,
                                2.60261209e-06,
                                2.56234446e-06,
                                2.11162563e-06,
                                1.61774868e-06,
                                1.39431387e-06,
                                1.64217088e-06,
                                2.31360784e-06,
                                3.13090619e-06,
                                3.83908225e-06,
                                4.35302731e-06,
                                4.67870504e-06,
                                4.84354637e-06,
                                4.90376869e-06,
                                4.91908875e-06,
                                4.92196879e-06,
                                4.92248720e-06,
                                4.92261931e-06,
                                4.92266441e-06,
                            ],
                            [
                                2.07300000e-06,
                                2.07300241e-06,
                                2.07365590e-06,
                                2.10782176e-06,
                                2.47023394e-06,
                                3.29595019e-06,
                                3.71184421e-06,
                                3.75625892e-06,
                                3.71789509e-06,
                                3.31607852e-06,
                                2.49871058e-06,
                                2.50161165e-06,
                                3.22503390e-06,
                                3.45199857e-06,
                                2.99394321e-06,
                                2.51456240e-06,
                                2.37387787e-06,
                                2.63750456e-06,
                                3.21284100e-06,
                                3.87176041e-06,
                                4.35613565e-06,
                                4.57984935e-06,
                                4.71330577e-06,
                                5.02541464e-06,
                                5.14208904e-06,
                                5.17195834e-06,
                                5.17758173e-06,
                                5.17859423e-06,
                                5.17885226e-06,
                                5.17894035e-06,
                            ],
                            [
                                2.07300000e-06,
                                2.07300241e-06,
                                2.07365590e-06,
                                2.10782176e-06,
                                2.47023394e-06,
                                3.29595019e-06,
                                3.71184421e-06,
                                3.75625892e-06,
                                3.71789509e-06,
                                3.31607852e-06,
                                2.49871058e-06,
                                2.50161165e-06,
                                3.22503390e-06,
                                3.45199857e-06,
                                2.99394321e-06,
                                2.51456240e-06,
                                2.37387787e-06,
                                2.63750456e-06,
                                3.21284100e-06,
                                3.87176041e-06,
                                4.35613565e-06,
                                4.57984935e-06,
                                4.71330577e-06,
                                5.02541464e-06,
                                5.14208904e-06,
                                5.17195834e-06,
                                5.17758173e-06,
                                5.17859423e-06,
                                5.17885226e-06,
                                5.17894035e-06,
                            ],
                        ]
                    ),
                ]
            ],
            reflectivityXData=[
                np.array(
                    [
                        [
                            0.011403,
                            0.013861,
                            0.016848,
                            0.020479,
                            0.024892,
                            0.030256,
                            0.036777,
                            0.044702,
                            0.054336,
                            0.066045,
                            0.080279,
                            0.097579,
                            0.118610,
                            0.144170,
                            0.175240,
                            0.213000,
                            0.258910,
                            0.314700,
                            0.382520,
                            0.464960,
                            0.565160,
                        ],
                    ],
                ),
                np.array(
                    [
                        [
                            0.011403,
                            0.013861,
                            0.016848,
                            0.020479,
                            0.024892,
                            0.030256,
                            0.036777,
                            0.044702,
                            0.054336,
                            0.066045,
                            0.080279,
                            0.097579,
                            0.118610,
                            0.144170,
                            0.175240,
                            0.213000,
                            0.258910,
                            0.314700,
                            0.382520,
                            0.464960,
                            0.565160,
                        ],
                    ],
                ),
            ],
            sldXData=[
                [
                    np.array(
                        [
                            [
                                0.0,
                                11.0,
                                22.0,
                                33.0,
                                44.0,
                                55.0,
                                66.0,
                                77.0,
                                88.0,
                                99.0,
                                110.0,
                                121.0,
                                132.0,
                                143.0,
                                154.0,
                                165.0,
                                176.0,
                                187.0,
                                198.0,
                                209.0,
                                220.0,
                                231.0,
                                242.0,
                                253.0,
                                264.0,
                                275.0,
                                286.0,
                                297.0,
                                308.0,
                                319.0,
                            ],
                        ],
                    ),
                    np.array(
                        [
                            [
                                0.0,
                                11.0,
                                22.0,
                                33.0,
                                44.0,
                                55.0,
                                66.0,
                                77.0,
                                88.0,
                                99.0,
                                110.0,
                                121.0,
                                132.0,
                                143.0,
                                154.0,
                                165.0,
                                176.0,
                                187.0,
                                198.0,
                                209.0,
                                220.0,
                                231.0,
                                242.0,
                                253.0,
                                264.0,
                                275.0,
                                286.0,
                                297.0,
                                308.0,
                                319.0,
                            ],
                        ],
                    ),
                ],
            ],
            sampleChi=np.array(
                [
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.46133559e16,
                    1.12748255e06,
                    1.12748255e06,
                    1.46133559e16,
                ],
            ),
        ),
        confidenceIntervals=RATapi.outputs.ConfidenceIntervals(
            percentile65=np.array(
                [
                    [
                        -1.29074586e-231,
                        8.33251318e000,
                        1.75397363e001,
                        1.75397363e001,
                        9.85302945e000,
                        9.85302945e000,
                        8.34197863e000,
                        8.34197863e000,
                        1.65750684e001,
                        1.45435510e001,
                        1.45435510e001,
                        1.52609047e001,
                        4.82866779e001,
                        2.91225389e001,
                        5.60206314e-006,
                        7.08098641e-007,
                        7.08098641e-007,
                        4.65378475e-006,
                    ],
                    [
                        8.33251318e000,
                        5.48185565e001,
                        5.48185565e001,
                        4.57554170e001,
                        4.57554170e001,
                        1.17557273e001,
                        1.17557273e001,
                        3.18752608e001,
                        3.18752608e001,
                        1.65750684e001,
                        1.52609047e001,
                        4.88237113e001,
                        4.88237113e001,
                        4.82866779e001,
                        2.91225389e001,
                        5.60206314e-006,
                        5.87958515e-006,
                        5.87958515e-006,
                    ],
                ],
            ),
            percentile95=np.array(
                [
                    [
                        -1.29074586e-231,
                        8.33251318e000,
                        1.75397363e001,
                        1.75397363e001,
                        9.85302945e000,
                        9.85302945e000,
                        8.34197863e000,
                        8.34197863e000,
                        1.65750684e001,
                        1.45435510e001,
                        1.45435510e001,
                        1.52609047e001,
                        4.82866779e001,
                        2.91225389e001,
                        5.60206314e-006,
                        7.08098641e-007,
                        7.08098641e-007,
                        4.65378475e-006,
                    ],
                    [
                        8.33251318e000,
                        5.48185565e001,
                        5.48185565e001,
                        4.57554170e001,
                        4.57554170e001,
                        1.17557273e001,
                        1.17557273e001,
                        3.18752608e001,
                        3.18752608e001,
                        1.65750684e001,
                        1.52609047e001,
                        4.88237113e001,
                        4.88237113e001,
                        4.82866779e001,
                        2.91225389e001,
                        5.60206314e-006,
                        5.87958515e-006,
                        5.87958515e-006,
                    ],
                ],
            ),
            mean=np.array(
                [
                    [
                        4.16625659e00,
                        3.15755349e01,
                        3.61791464e01,
                        3.16475766e01,
                        2.78042232e01,
                        1.08043784e01,
                        1.00488530e01,
                        2.01086197e01,
                        2.42251646e01,
                        1.55593097e01,
                        1.49022278e01,
                        3.20423080e01,
                        4.85551946e01,
                        3.87046084e01,
                        1.45612723e01,
                        3.15508089e-06,
                        3.29384190e-06,
                        5.26668495e-06,
                    ],
                ],
            ),
        ),
        dreamParams=RATapi.outputs.DreamParams(
            nParams=18.0,
            nChains=1.0,
            nGenerations=1.0,
            parallel=False,
            CPU=1.0,
            jumpProbability=0.5,
            pUnitGamma=0.2,
            nCR=3.0,
            delta=3.0,
            steps=50.0,
            zeta=1e-12,
            outlier="iqr",
            adaptPCR=False,
            thinning=1.0,
            epsilon=0.025,
            ABC=False,
            IO=False,
            storeOutput=False,
            R=np.array([[0.0]]),
        ),
        dreamOutput=RATapi.outputs.DreamOutput(
            allChains=np.array(
                [
                    [
                        [8.33251318e00],
                        [5.48185565e01],
                        [1.75397363e01],
                        [4.57554170e01],
                        [9.85302945e00],
                        [1.17557273e01],
                        [8.34197863e00],
                        [3.18752608e01],
                        [1.65750684e01],
                        [1.45435510e01],
                        [1.52609047e01],
                        [4.88237113e01],
                        [4.82866779e01],
                        [2.91225389e01],
                        [5.60206314e-06],
                        [7.08098641e-07],
                        [5.87958515e-06],
                        [4.65378475e-06],
                        [-4.86509830e01],
                        [-5.63741277e05],
                    ],
                ],
            ),
            outlierChains=np.array([[0.0, 0.0]]),
            runtime=2.6e-06,
            iteration=2.0,
            modelOutput=0.0,
            AR=np.array([[1.0, np.nan]]),
            R_stat=np.array(
                [
                    [
                        1.0,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                        np.nan,
                    ],
                ],
            ),
            CR=np.array([[1.0, 0.33333333, 0.33333333, 0.33333333]]),
        ),
        nestedSamplerOutput=RATapi.outputs.NestedSamplerOutput(
            logZ=0.0,
            nestSamples=np.array([[0.0, 0.0]]),
            postSamples=np.array([[0.0, 0.0]]),
        ),
        chain=np.array(
            [
                [
                    -1.29231905e-231,
                    8.33251318e000,
                    5.48185565e001,
                    1.75397363e001,
                    4.57554170e001,
                    9.85302945e000,
                    1.17557273e001,
                    8.34197863e000,
                    3.18752608e001,
                    1.65750684e001,
                    1.45435510e001,
                    1.52609047e001,
                    4.88237113e001,
                    4.82866779e001,
                    2.91225389e001,
                    5.60206314e-006,
                    7.08098641e-007,
                    5.87958515e-006,
                ],
                [
                    8.33251318e000,
                    5.48185565e001,
                    1.75397363e001,
                    4.57554170e001,
                    9.85302945e000,
                    1.17557273e001,
                    8.34197863e000,
                    3.18752608e001,
                    1.65750684e001,
                    1.45435510e001,
                    1.52609047e001,
                    4.88237113e001,
                    4.82866779e001,
                    2.91225389e001,
                    5.60206314e-006,
                    7.08098641e-007,
                    5.87958515e-006,
                    4.65378475e-006,
                ],
            ],
        ),
    )


@pytest.fixture
def r1_default_project():
    """The Project corresponding to the data in R1defaultProject.mat."""

    project = RATapi.Project(
        name="defaultProject",
        calculation="non polarised",
        model="standard layers",
        geometry="air/substrate",
        absorption=False,
        parameters=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Substrate Roughness",
                min=3.0,
                value=4.844363132849221,
                max=8.0,
                fit=True,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        bulk_in=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Air",
                min=0.0,
                value=0.0,
                max=0.0,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        bulk_out=RATapi.ClassList(
            RATapi.models.Parameter(
                name="D2O",
                min=6.3e-06,
                value=6.35e-06,
                max=6.4e-06,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        scalefactors=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Scalefactor 1",
                min=0.009999999776482582,
                value=0.10141560336360426,
                max=0.3,
                fit=True,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        backgrounds=RATapi.ClassList(
            RATapi.models.Background(name="Background 1", type="constant", value_1="Background parameter 1")
        ),
        background_parameters=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Background parameter 1",
                min=5e-08,
                value=3.069003361230152e-06,
                max=7e-06,
                fit=True,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        resolutions=RATapi.ClassList(
            RATapi.models.Resolution(name="Resolution 1", type="constant", value_1="Resolution parameter 1")
        ),
        resolution_parameters=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Resolution parameter 1",
                min=0.01,
                value=0.03,
                max=0.05,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
    )
    project.data.append(
        name="f82395c",
        data=np.array(
            [
                [4.8866e-02, 1.2343e-04, 1.3213e-06],
                [5.1309e-02, 1.0063e-04, 1.0803e-06],
                [5.3874e-02, 8.2165e-05, 8.8779e-07],
                [5.6568e-02, 6.4993e-05, 7.2018e-07],
                [5.9396e-02, 5.3958e-05, 6.0015e-07],
                [6.2366e-02, 4.3590e-05, 5.0129e-07],
                [6.5485e-02, 3.5780e-05, 4.1957e-07],
                [6.8759e-02, 2.9130e-05, 3.5171e-07],
                [7.2197e-02, 2.3481e-05, 3.0586e-07],
                [7.5807e-02, 1.8906e-05, 2.6344e-07],
                [7.9597e-02, 1.4642e-05, 2.2314e-07],
                [8.3577e-02, 1.1589e-05, 1.8938e-07],
                [8.7756e-02, 9.5418e-06, 1.6220e-07],
                [9.2143e-02, 7.5694e-06, 1.3809e-07],
                [9.6751e-02, 6.3831e-06, 1.2097e-07],
                [1.0159e-01, 5.0708e-06, 1.0333e-07],
                [1.0667e-01, 4.1041e-06, 8.9548e-08],
                [1.1200e-01, 3.4253e-06, 7.9830e-08],
                [1.1760e-01, 2.8116e-06, 7.1554e-08],
                [1.2348e-01, 2.3767e-06, 6.3738e-08],
                [1.2966e-01, 1.9241e-06, 5.6586e-08],
                [1.3614e-01, 1.5642e-06, 5.2778e-08],
                [1.4294e-01, 1.2922e-06, 4.9730e-08],
                [1.5009e-01, 1.1694e-06, 5.1175e-08],
                [1.5760e-01, 9.7837e-07, 5.0755e-08],
                [1.6548e-01, 8.9138e-07, 5.3542e-08],
                [1.7375e-01, 7.9420e-07, 5.4857e-08],
                [1.8244e-01, 7.9131e-07, 5.8067e-08],
                [1.9156e-01, 6.5358e-07, 5.7717e-08],
                [2.0114e-01, 6.2970e-07, 5.7951e-08],
                [2.1119e-01, 5.0130e-07, 5.5262e-08],
                [2.2175e-01, 5.0218e-07, 5.6461e-08],
                [2.3284e-01, 3.9299e-07, 5.0685e-08],
                [2.4448e-01, 3.5324e-07, 5.0194e-08],
                [2.5671e-01, 4.4475e-07, 5.6485e-08],
                [2.6954e-01, 5.1338e-07, 6.2247e-08],
                [2.8302e-01, 3.4918e-07, 4.9745e-08],
                [2.9717e-01, 4.3037e-07, 5.5488e-08],
                [3.1203e-01, 4.0099e-07, 5.3591e-08],
                [3.2763e-01, 3.8397e-07, 5.1303e-08],
                [3.4401e-01, 3.0995e-07, 4.5965e-08],
                [3.6121e-01, 3.9357e-07, 5.0135e-08],
                [3.7927e-01, 3.0997e-07, 4.3680e-08],
                [3.9824e-01, 2.9656e-07, 4.2432e-08],
                [4.1815e-01, 2.1909e-07, 3.6117e-08],
                [4.3906e-01, 2.3153e-07, 3.6307e-08],
                [4.6101e-01, 3.3428e-07, 4.3874e-08],
                [4.8406e-01, 2.3441e-07, 3.7488e-08],
                [5.0826e-01, 1.5496e-07, 3.0585e-08],
                [5.3368e-01, 2.4708e-07, 3.9376e-08],
                [5.6036e-01, 2.2157e-07, 3.8258e-08],
                [5.8838e-01, 2.2798e-07, 4.6976e-08],
                [6.1169e-01, 6.0272e-07, 2.3239e-07],
            ]
        ),
        data_range=[0.048866, 0.61169],
        simulation_range=[0.048866, 0.61169],
    )
    project.contrasts.append(
        name="Chain-d, acmw",
        data="f82395c",
        background="Background 1",
        background_action="add",
        bulk_in="Air",
        bulk_out="D2O",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=[],
    )
    return project


@pytest.fixture
def r1_monolayer():
    """The Project file corresponding to the data in R1monolayerVolumeModel.mat."""
    project = RATapi.Project(
        name="monolayerVolumeModel",
        calculation="non polarised",
        model="custom layers",
        geometry="air/substrate",
        absorption=False,
        parameters=RATapi.ClassList(
            [
                RATapi.models.Parameter(
                    name="Substrate Roughness",
                    min=1.0,
                    value=2.9979642781948908,
                    max=8.0,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
                RATapi.models.Parameter(
                    name="Area per molecule",
                    min=47.0,
                    value=53.052680457664785,
                    max=100.0,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
                RATapi.models.Parameter(
                    name="Head Thickness",
                    min=7.0,
                    value=12.276333836779942,
                    max=20.0,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
                RATapi.models.Parameter(
                    name="Theta",
                    min=0.0,
                    value=28.870541049836262,
                    max=50.0,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
            ]
        ),
        bulk_in=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Air",
                min=0.0,
                value=0.0,
                max=0.0,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        bulk_out=(
            RATapi.ClassList(
                [
                    RATapi.models.Parameter(
                        name="D2O",
                        min=6.3e-06,
                        value=6.35e-06,
                        max=6.4e-06,
                        fit=False,
                        prior_type="uniform",
                        mu=0.0,
                        sigma=np.inf,
                    ),
                    RATapi.models.Parameter(
                        name="ACMW",
                        min=-5e-07,
                        value=0.0,
                        max=5e-07,
                        fit=False,
                        prior_type="uniform",
                        mu=0.0,
                        sigma=np.inf,
                    ),
                ]
            )
        ),
        scalefactors=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Scalefactor 1",
                min=0.1,
                value=0.2272676786810902,
                max=0.4,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        backgrounds=RATapi.ClassList(
            [
                RATapi.models.Background(name="Background  D2O", type="constant", value_1="Background parameter 1"),
                RATapi.models.Background(name="Background ACMW", type="constant", value_1="Background parameter 2"),
            ]
        ),
        background_parameters=RATapi.ClassList(
            [
                RATapi.models.Parameter(
                    name="Background parameter 1",
                    min=1e-07,
                    value=2.2653463958223856e-06,
                    max=7e-06,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
                RATapi.models.Parameter(
                    name="Background parameter 2",
                    min=1e-07,
                    value=5.7431759430575025e-06,
                    max=7e-06,
                    fit=True,
                    prior_type="uniform",
                    mu=0.0,
                    sigma=np.inf,
                ),
            ]
        ),
        resolutions=RATapi.ClassList(
            RATapi.models.Resolution(name="Resolution 1", type="constant", value_1="Resolution parameter 1")
        ),
        resolution_parameters=RATapi.ClassList(
            RATapi.models.Parameter(
                name="Resolution parameter 1",
                min=0.01,
                value=0.03,
                max=0.05,
                fit=False,
                prior_type="uniform",
                mu=0.0,
                sigma=np.inf,
            )
        ),
        custom_files=RATapi.ClassList(
            RATapi.models.CustomFile(
                name="Model_IIb", filename="Model_IIb.m", function_name="Model_IIb", language="matlab", path=""
            )
        ),
    )

    project.data.append(
        name="d70acmw20",
        data=np.array(
            [
                [5.1793e-02, 2.3792e-04, 3.6498e-06],
                [5.4383e-02, 2.2030e-04, 3.1361e-06],
                [5.7102e-02, 1.9302e-04, 2.6450e-06],
                [5.9957e-02, 1.7223e-04, 2.2673e-06],
                [6.2955e-02, 1.5210e-04, 1.9699e-06],
                [6.6103e-02, 1.3650e-04, 1.7055e-06],
                [6.9408e-02, 1.1806e-04, 1.5072e-06],
                [7.2878e-02, 1.0904e-04, 1.3658e-06],
                [7.6522e-02, 9.5873e-05, 1.2024e-06],
                [8.0348e-02, 8.2201e-05, 1.0754e-06],
                [8.4365e-02, 7.0764e-05, 9.4576e-07],
                [8.8584e-02, 6.2973e-05, 8.3975e-07],
                [9.3013e-02, 5.3710e-05, 7.3854e-07],
                [9.7664e-02, 4.6651e-05, 6.5893e-07],
                [1.0255e-01, 4.0448e-05, 5.8945e-07],
                [1.0767e-01, 3.4559e-05, 5.2655e-07],
                [1.1306e-01, 2.9817e-05, 4.7806e-07],
                [1.1871e-01, 2.5965e-05, 4.4165e-07],
                [1.2465e-01, 2.1963e-05, 3.9117e-07],
                [1.3088e-01, 1.8535e-05, 3.5668e-07],
                [1.3742e-01, 1.6299e-05, 3.4665e-07],
                [1.4429e-01, 1.3704e-05, 3.2491e-07],
                [1.5151e-01, 1.0571e-05, 2.9830e-07],
                [1.5908e-01, 8.9320e-06, 2.9188e-07],
                [1.6704e-01, 6.8328e-06, 2.7922e-07],
                [1.7539e-01, 5.7856e-06, 2.7261e-07],
                [1.8416e-01, 4.5358e-06, 2.5335e-07],
                [1.9337e-01, 3.9935e-06, 2.5336e-07],
                [2.0304e-01, 3.1800e-06, 5.2440e-07],
                [2.1319e-01, 2.7006e-06, 4.6690e-07],
                [2.2385e-01, 2.2210e-06, 4.5090e-07],
                [2.3504e-01, 1.6163e-06, 4.9340e-07],
                [2.4679e-01, 1.7554e-06, 3.2120e-07],
                [2.5913e-01, 1.3194e-06, 1.7170e-07],
                [2.7209e-01, 1.5033e-06, 3.4760e-07],
                [2.8569e-01, 1.3801e-06, 7.2017e-07],
                [2.9998e-01, 1.2640e-06, 4.5486e-07],
                [3.1497e-01, 1.0585e-06, 3.6820e-07],
                [3.3072e-01, 1.3063e-06, 4.5401e-07],
                [3.4726e-01, 1.1229e-06, 2.3240e-07],
                [3.6462e-01, 1.5197e-06, 4.8780e-07],
                [3.8285e-01, 1.1419e-06, 4.1880e-07],
                [4.0200e-01, 1.3285e-06, 4.0215e-07],
                [4.2210e-01, 1.2467e-06, 5.8010e-07],
                [4.4320e-01, 1.1322e-06, 3.2730e-07],
                [4.6536e-01, 1.1336e-06, 4.3291e-07],
                [4.8863e-01, 1.4843e-06, 5.0110e-07],
                [5.1306e-01, 1.2055e-06, 3.0660e-07],
                [5.3871e-01, 1.2170e-06, 4.0500e-07],
                [5.6565e-01, 8.9674e-07, 3.5866e-07],
                [5.8877e-01, 9.4625e-07, 7.6511e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )

    project.data.append(
        name="d70d2o20",
        data=np.array(
            [
                [5.1793e-02, 1.4943e-04, 3.2517e-06],
                [5.4383e-02, 1.1882e-04, 2.5846e-06],
                [5.7102e-02, 9.2164e-05, 2.0507e-06],
                [5.9957e-02, 7.7813e-05, 1.7070e-06],
                [6.2955e-02, 6.1143e-05, 1.3983e-06],
                [6.6103e-02, 4.8033e-05, 1.1343e-06],
                [6.9408e-02, 4.1379e-05, 1.0006e-06],
                [7.2878e-02, 3.5090e-05, 8.6919e-07],
                [7.6522e-02, 3.0350e-05, 7.5890e-07],
                [8.0348e-02, 2.4115e-05, 6.5226e-07],
                [8.4365e-02, 2.0755e-05, 5.7445e-07],
                [8.8584e-02, 1.7500e-05, 4.9617e-07],
                [9.3013e-02, 1.5011e-05, 4.3754e-07],
                [9.7664e-02, 1.3632e-05, 3.9907e-07],
                [1.0255e-01, 1.2997e-05, 3.7469e-07],
                [1.0767e-01, 1.1656e-05, 3.4282e-07],
                [1.1306e-01, 1.0820e-05, 3.2299e-07],
                [1.1871e-01, 9.5831e-06, 3.0088e-07],
                [1.2465e-01, 8.9996e-06, 2.8102e-07],
                [1.3088e-01, 8.6498e-06, 2.7363e-07],
                [1.3742e-01, 7.9412e-06, 2.7158e-07],
                [1.4429e-01, 7.4042e-06, 2.6829e-07],
                [1.5151e-01, 6.8321e-06, 2.6877e-07],
                [1.5908e-01, 5.6809e-06, 2.6232e-07],
                [1.6704e-01, 5.6646e-06, 2.8607e-07],
                [1.7539e-01, 4.7762e-06, 2.7767e-07],
                [1.8416e-01, 4.1971e-06, 2.7353e-07],
                [1.9337e-01, 4.1815e-06, 2.9140e-07],
                [2.0304e-01, 3.2725e-06, 2.3160e-07],
                [2.1319e-01, 2.3244e-06, 4.2270e-07],
                [2.2385e-01, 2.2892e-06, 1.4970e-07],
                [2.3504e-01, 1.9198e-06, 8.1460e-07],
                [2.4679e-01, 1.4828e-06, 1.0840e-07],
                [2.5913e-01, 1.1450e-06, 4.7440e-07],
                [2.7209e-01, 1.3047e-06, 1.8840e-07],
                [2.8569e-01, 9.6081e-07, 3.9385e-07],
                [2.9998e-01, 8.2280e-07, 2.3955e-07],
                [3.1497e-01, 6.3219e-07, 3.5324e-07],
                [3.3072e-01, 6.1254e-07, 5.0846e-07],
                [3.4726e-01, 7.4092e-07, 2.2312e-07],
                [3.6462e-01, 6.3584e-07, 4.2866e-07],
                [3.8285e-01, 7.7287e-07, 2.9493e-07],
                [4.0200e-01, 9.4637e-07, 2.3347e-07],
                [4.2210e-01, 7.0576e-07, 3.3494e-07],
                [4.4320e-01, 7.8969e-07, 2.3371e-07],
                [4.6536e-01, 5.8374e-07, 3.2624e-07],
                [4.8863e-01, 5.6711e-07, 6.0419e-07],
                [5.1306e-01, 4.7782e-07, 3.4822e-07],
                [5.3871e-01, 6.3813e-07, 4.3279e-07],
                [5.6565e-01, 4.6186e-07, 1.8863e-07],
                [5.8877e-01, 5.6129e-07, 4.3960e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )
    project.data.append(
        name="d13acmw20",
        data=np.array(
            [
                [5.1793e-02, 1.8087e-05, 7.9683e-07],
                [5.4383e-02, 1.4582e-05, 6.3691e-07],
                [5.7102e-02, 1.3621e-05, 5.5407e-07],
                [5.9957e-02, 1.2409e-05, 4.8014e-07],
                [6.2955e-02, 1.0992e-05, 4.1824e-07],
                [6.6103e-02, 1.0619e-05, 3.7541e-07],
                [6.9408e-02, 1.0162e-05, 3.4945e-07],
                [7.2878e-02, 1.0716e-05, 3.3763e-07],
                [7.6522e-02, 8.3468e-06, 2.8040e-07],
                [8.0348e-02, 7.5250e-06, 2.5702e-07],
                [8.4365e-02, 6.4395e-06, 2.2528e-07],
                [8.8584e-02, 6.0544e-06, 2.0545e-07],
                [9.3013e-02, 5.9517e-06, 1.9417e-07],
                [9.7664e-02, 5.6433e-06, 1.8108e-07],
                [1.0255e-01, 4.8172e-06, 1.6076e-07],
                [1.0767e-01, 4.8066e-06, 1.5520e-07],
                [1.1306e-01, 4.1559e-06, 1.4115e-07],
                [1.1871e-01, 4.1418e-06, 1.3926e-07],
                [1.2465e-01, 3.4580e-06, 1.2273e-07],
                [1.3088e-01, 3.2376e-06, 1.1791e-07],
                [1.3742e-01, 3.0976e-06, 1.1948e-07],
                [1.4429e-01, 2.7478e-06, 1.1518e-07],
                [1.5151e-01, 2.5492e-06, 1.1573e-07],
                [1.5908e-01, 2.7500e-06, 1.2812e-07],
                [1.6704e-01, 2.1813e-06, 2.0450e-07],
                [1.7539e-01, 1.8609e-06, 5.4900e-07],
                [1.8416e-01, 1.9405e-06, 1.5660e-07],
                [1.9337e-01, 1.7421e-06, 2.3280e-07],
                [2.0304e-01, 1.8050e-06, 3.9820e-07],
                [2.1319e-01, 1.5801e-06, 1.4110e-07],
                [2.2385e-01, 1.6724e-06, 7.7900e-08],
                [2.3504e-01, 1.4150e-06, 4.0820e-07],
                [2.4679e-01, 1.4340e-06, 5.3180e-07],
                [2.5913e-01, 1.3102e-06, 2.6000e-07],
                [2.7209e-01, 1.3702e-06, 2.8540e-07],
                [2.8569e-01, 1.2468e-06, 2.3230e-07],
                [2.9998e-01, 1.2956e-06, 5.3240e-07],
                [3.1497e-01, 1.2947e-06, 3.9940e-07],
                [3.3072e-01, 1.2488e-06, 2.1390e-07],
                [3.4726e-01, 1.4620e-06, 3.3640e-07],
                [3.6462e-01, 1.3056e-06, 2.1600e-07],
                [3.8285e-01, 1.4553e-06, 2.3170e-07],
                [4.0200e-01, 1.1579e-06, 2.6740e-07],
                [4.2210e-01, 1.1753e-06, 3.0940e-07],
                [4.4320e-01, 1.0066e-06, 5.2040e-07],
                [4.6536e-01, 1.1043e-06, 3.1870e-07],
                [4.8863e-01, 1.2969e-06, 4.1670e-07],
                [5.1306e-01, 1.2495e-06, 2.0890e-07],
                [5.3871e-01, 1.3898e-06, 4.7560e-07],
                [5.6565e-01, 1.1225e-06, 3.0470e-07],
                [5.8877e-01, 8.3346e-07, 3.8944e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )
    project.data.append(
        name="d13d2o20",
        data=np.array(
            [
                [5.1793e-02, 2.2464e-04, 3.4007e-06],
                [5.4383e-02, 1.8531e-04, 2.7523e-06],
                [5.7102e-02, 1.5463e-04, 2.2639e-06],
                [5.9957e-02, 1.2459e-04, 1.8443e-06],
                [6.2955e-02, 1.0332e-04, 1.5526e-06],
                [6.6103e-02, 8.3207e-05, 1.2736e-06],
                [6.9408e-02, 6.6859e-05, 1.0850e-06],
                [7.2878e-02, 5.5540e-05, 9.3212e-07],
                [7.6522e-02, 4.6869e-05, 8.0410e-07],
                [8.0348e-02, 3.6672e-05, 6.8717e-07],
                [8.4365e-02, 3.0955e-05, 5.9837e-07],
                [8.8584e-02, 2.4781e-05, 5.0395e-07],
                [9.3013e-02, 2.0163e-05, 4.3266e-07],
                [9.7664e-02, 1.7194e-05, 3.8276e-07],
                [1.0255e-01, 1.3801e-05, 3.2953e-07],
                [1.0767e-01, 1.0868e-05, 2.8268e-07],
                [1.1306e-01, 8.9118e-06, 2.5019e-07],
                [1.1871e-01, 8.2266e-06, 2.3789e-07],
                [1.2465e-01, 6.7891e-06, 2.0822e-07],
                [1.3088e-01, 5.5048e-06, 1.8610e-07],
                [1.3742e-01, 4.3011e-06, 1.7064e-07],
                [1.4429e-01, 3.7549e-06, 1.6291e-07],
                [1.5151e-01, 3.0281e-06, 1.5298e-07],
                [1.5908e-01, 2.7102e-06, 1.5371e-07],
                [1.6704e-01, 1.9817e-06, 5.2560e-07],
                [1.7539e-01, 1.8156e-06, 6.2930e-07],
                [1.8416e-01, 1.6126e-06, 1.0180e-07],
                [1.9337e-01, 1.5010e-06, 2.6400e-07],
                [2.0304e-01, 1.3504e-06, 3.6310e-07],
                [2.1319e-01, 1.0658e-06, 6.1751e-07],
                [2.2385e-01, 1.0028e-06, 2.4871e-07],
                [2.3504e-01, 9.1701e-07, 2.3499e-07],
                [2.4679e-01, 7.5092e-07, 2.4518e-07],
                [2.5913e-01, 1.0126e-06, 1.6429e-07],
                [2.7209e-01, 7.5725e-07, 4.1481e-07],
                [2.8569e-01, 6.2167e-07, 3.5735e-07],
                [2.9998e-01, 7.0777e-07, 2.0772e-07],
                [3.1497e-01, 8.5576e-07, 1.6916e-07],
                [3.3072e-01, 8.4284e-07, 2.2151e-07],
                [3.4726e-01, 4.4441e-07, 3.2246e-07],
                [3.6462e-01, 6.6378e-07, 2.3675e-07],
                [3.8285e-01, 7.1382e-07, 3.7039e-07],
                [4.0200e-01, 5.6163e-07, 4.4157e-07],
                [4.2210e-01, 5.9761e-07, 3.4927e-07],
                [4.4320e-01, 4.9182e-07, 3.5155e-07],
                [4.6536e-01, 5.6609e-07, 2.3558e-07],
                [4.8863e-01, 5.3547e-07, 1.7267e-07],
                [5.1306e-01, 4.7647e-07, 5.4127e-07],
                [5.3871e-01, 3.8841e-07, 2.9179e-07],
                [5.6565e-01, 4.7749e-07, 8.6240e-08],
                [5.8877e-01, 4.6278e-07, 2.3214e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )
    project.data.append(
        name="d83acmw20",
        data=np.array(
            [
                [5.1793e-02, 3.1675e-04, 4.7735e-06],
                [5.4383e-02, 2.7591e-04, 3.9806e-06],
                [5.7102e-02, 2.4370e-04, 3.3680e-06],
                [5.9957e-02, 2.1836e-04, 2.8915e-06],
                [6.2955e-02, 1.9492e-04, 2.5226e-06],
                [6.6103e-02, 1.7128e-04, 2.1638e-06],
                [6.9408e-02, 1.5381e-04, 1.9499e-06],
                [7.2878e-02, 1.3398e-04, 1.7140e-06],
                [7.6522e-02, 1.1643e-04, 1.5011e-06],
                [8.0348e-02, 9.7386e-05, 1.3256e-06],
                [8.4365e-02, 8.5091e-05, 1.1744e-06],
                [8.8584e-02, 7.2996e-05, 1.0224e-06],
                [9.3013e-02, 6.0693e-05, 8.8816e-07],
                [9.7664e-02, 5.1967e-05, 7.8767e-07],
                [1.0255e-01, 4.3371e-05, 6.9118e-07],
                [1.0767e-01, 3.6363e-05, 6.1123e-07],
                [1.1306e-01, 2.9948e-05, 5.4231e-07],
                [1.1871e-01, 2.5863e-05, 4.9879e-07],
                [1.2465e-01, 2.1196e-05, 4.3527e-07],
                [1.3088e-01, 1.7278e-05, 3.8999e-07],
                [1.3742e-01, 1.3647e-05, 3.5922e-07],
                [1.4429e-01, 1.0898e-05, 3.2810e-07],
                [1.5151e-01, 8.3399e-06, 2.9972e-07],
                [1.5908e-01, 6.7776e-06, 2.8761e-07],
                [1.6704e-01, 5.8668e-06, 2.9311e-07],
                [1.7539e-01, 4.1380e-06, 2.6133e-07],
                [1.8416e-01, 2.7663e-06, 2.2329e-07],
                [1.9337e-01, 2.3200e-06, 2.1903e-07],
                [2.0304e-01, 1.9127e-06, 5.6910e-07],
                [2.1319e-01, 1.7434e-06, 1.2640e-07],
                [2.2385e-01, 1.7550e-06, 1.9990e-07],
                [2.3504e-01, 1.4527e-06, 3.8840e-07],
                [2.4679e-01, 1.1333e-06, 5.9440e-07],
                [2.5913e-01, 1.1667e-06, 5.7100e-07],
                [2.7209e-01, 1.3539e-06, 2.1560e-07],
                [2.8569e-01, 1.5357e-06, 3.5990e-07],
                [2.9998e-01, 9.0768e-07, 8.9822e-07],
                [3.1497e-01, 1.2821e-06, 3.1450e-07],
                [3.3072e-01, 1.2360e-06, 4.1700e-07],
                [3.4726e-01, 9.6932e-07, 4.5458e-07],
                [3.6462e-01, 1.4644e-06, 4.7300e-07],
                [3.8285e-01, 1.2330e-06, 3.4780e-07],
                [4.0200e-01, 1.3783e-06, 4.5143e-07],
                [4.2210e-01, 1.2637e-06, 2.4980e-07],
                [4.4320e-01, 1.1299e-06, 1.9310e-07],
                [4.6536e-01, 1.1748e-06, 4.3340e-07],
                [4.8863e-01, 1.2029e-06, 1.4040e-07],
                [5.1306e-01, 9.9648e-07, 6.1262e-07],
                [5.3871e-01, 9.4235e-07, 6.9825e-07],
                [5.6565e-01, 1.2561e-06, 4.5650e-07],
                [5.8877e-01, 1.0005e-06, 1.1520e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )
    project.data.append(
        name="d83d2o20",
        data=np.array(
            [
                [5.1793e-02, 2.3497e-04, 4.1366e-06],
                [5.4383e-02, 1.9138e-04, 3.3298e-06],
                [5.7102e-02, 1.6092e-04, 2.7523e-06],
                [5.9957e-02, 1.3177e-04, 2.2579e-06],
                [6.2955e-02, 1.0694e-04, 1.8785e-06],
                [6.6103e-02, 9.0449e-05, 1.5805e-06],
                [6.9408e-02, 7.2115e-05, 1.3407e-06],
                [7.2878e-02, 5.9602e-05, 1.1494e-06],
                [7.6522e-02, 4.9097e-05, 9.7903e-07],
                [8.0348e-02, 4.1264e-05, 8.6633e-07],
                [8.4365e-02, 3.2611e-05, 7.3101e-07],
                [8.8584e-02, 2.7676e-05, 6.3311e-07],
                [9.3013e-02, 2.2224e-05, 5.4031e-07],
                [9.7664e-02, 1.9035e-05, 4.7882e-07],
                [1.0255e-01, 1.5324e-05, 4.1267e-07],
                [1.0767e-01, 1.3008e-05, 3.6731e-07],
                [1.1306e-01, 1.0121e-05, 3.1672e-07],
                [1.1871e-01, 9.3837e-06, 3.0201e-07],
                [1.2465e-01, 7.4003e-06, 2.5840e-07],
                [1.3088e-01, 6.6417e-06, 2.4288e-07],
                [1.3742e-01, 5.5185e-06, 2.2950e-07],
                [1.4429e-01, 4.6301e-06, 2.1511e-07],
                [1.5151e-01, 3.8179e-06, 2.0421e-07],
                [1.5908e-01, 3.6504e-06, 2.1327e-07],
                [1.6704e-01, 3.3693e-06, 5.8910e-07],
                [1.7539e-01, 2.9089e-06, 6.0910e-07],
                [1.8416e-01, 2.0826e-06, 8.8760e-07],
                [1.9337e-01, 1.8413e-06, 2.0110e-07],
                [2.0304e-01, 1.7280e-06, 4.2860e-07],
                [2.1319e-01, 2.0794e-06, 4.5100e-07],
                [2.2385e-01, 1.2345e-06, 4.2410e-07],
                [2.3504e-01, 1.3707e-06, 2.7770e-07],
                [2.4679e-01, 1.2666e-06, 9.8800e-08],
                [2.5913e-01, 1.1391e-06, 6.9332e-07],
                [2.7209e-01, 1.1299e-06, 2.1680e-07],
                [2.8569e-01, 8.4537e-07, 5.3956e-07],
                [2.9998e-01, 6.8579e-07, 4.6611e-07],
                [3.1497e-01, 8.3948e-07, 3.2138e-07],
                [3.3072e-01, 8.8276e-07, 4.6705e-07],
                [3.4726e-01, 7.8563e-07, 6.4686e-07],
                [3.6462e-01, 6.8788e-07, 3.9930e-07],
                [3.8285e-01, 6.1412e-07, 1.8886e-07],
                [4.0200e-01, 6.8538e-07, 4.0952e-07],
                [4.2210e-01, 5.8111e-07, 5.0615e-07],
                [4.4320e-01, 5.4626e-07, 2.3667e-07],
                [4.6536e-01, 6.5234e-07, 4.2512e-07],
                [4.8863e-01, 3.6587e-07, 4.3870e-07],
                [5.1306e-01, 4.1699e-07, 1.9893e-07],
                [5.3871e-01, 4.9769e-07, 2.1907e-07],
                [5.6565e-01, 3.7836e-07, 3.9193e-07],
                [5.8877e-01, 4.2082e-07, 3.5255e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )
    project.data.append(
        name="hd2o20",
        data=np.array(
            [
                [5.1793e-02, 2.1311e-04, 3.8170e-06],
                [5.4383e-02, 1.7454e-04, 3.0814e-06],
                [5.7102e-02, 1.4351e-04, 2.5155e-06],
                [5.9957e-02, 1.1515e-04, 2.0424e-06],
                [6.2955e-02, 9.0901e-05, 1.6762e-06],
                [6.6103e-02, 7.4677e-05, 1.3909e-06],
                [6.9408e-02, 6.1965e-05, 1.2044e-06],
                [7.2878e-02, 4.9933e-05, 1.0186e-06],
                [7.6522e-02, 4.1322e-05, 8.7025e-07],
                [8.0348e-02, 3.2145e-05, 7.4114e-07],
                [8.4365e-02, 2.6927e-05, 6.4343e-07],
                [8.8584e-02, 2.0970e-05, 5.3415e-07],
                [9.3013e-02, 1.6109e-05, 4.4582e-07],
                [9.7664e-02, 1.2974e-05, 3.8305e-07],
                [1.0255e-01, 1.0411e-05, 3.2993e-07],
                [1.0767e-01, 7.7586e-06, 2.7512e-07],
                [1.1306e-01, 6.8947e-06, 2.5357e-07],
                [1.1871e-01, 5.5102e-06, 2.2445e-07],
                [1.2465e-01, 4.3486e-06, 1.9200e-07],
                [1.3088e-01, 3.3433e-06, 1.6724e-07],
                [1.3742e-01, 2.7268e-06, 1.5634e-07],
                [1.4429e-01, 2.2001e-06, 1.4377e-07],
                [1.5151e-01, 1.8064e-06, 1.3591e-07],
                [1.5908e-01, 1.4947e-06, 1.3214e-07],
                [1.6704e-01, 1.5335e-06, 3.9200e-07],
                [1.7539e-01, 1.0430e-06, 8.5200e-08],
                [1.8416e-01, 1.0070e-06, 3.9617e-07],
                [1.9337e-01, 7.5452e-07, 1.7785e-07],
                [2.0304e-01, 6.9356e-07, 2.8065e-07],
                [2.1319e-01, 5.9018e-07, 4.0612e-07],
                [2.2385e-01, 7.0286e-07, 3.4384e-07],
                [2.3504e-01, 6.0890e-07, 2.8376e-07],
                [2.4679e-01, 8.0600e-07, 2.2092e-07],
                [2.5913e-01, 7.9177e-07, 3.4918e-07],
                [2.7209e-01, 8.5193e-07, 3.3083e-07],
                [2.8569e-01, 8.1507e-07, 2.7682e-07],
                [2.9998e-01, 7.1477e-07, 1.6790e-07],
                [3.1497e-01, 6.1835e-07, 2.0612e-07],
                [3.3072e-01, 5.7863e-07, 3.3869e-07],
                [3.4726e-01, 4.0735e-07, 6.1275e-07],
                [3.6462e-01, 8.2937e-07, 3.2510e-07],
                [3.8285e-01, 7.5561e-07, 2.0851e-07],
                [4.0200e-01, 6.6174e-07, 2.9566e-07],
                [4.2210e-01, 5.6503e-07, 2.0503e-07],
                [4.4320e-01, 6.5599e-07, 3.0025e-07],
                [4.6536e-01, 4.8085e-07, 2.6812e-07],
                [4.8863e-01, 4.1179e-07, 2.7538e-07],
                [5.1306e-01, 6.0006e-07, 2.6417e-07],
                [5.3871e-01, 6.3395e-07, 3.9303e-07],
                [5.6565e-01, 6.3474e-07, 2.3217e-07],
                [5.8877e-01, 3.3414e-07, 3.9170e-07],
            ]
        ),
        data_range=[0.051793, 0.58877],
        simulation_range=[0.051793, 0.58877],
    )

    project.contrasts.append(
        name="d70, acmw",
        data="d70acmw20",
        background="Background ACMW",
        background_action="add",
        bulk_in="Air",
        bulk_out="ACMW",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="d70 d2o",
        data="d70d2o20",
        background="Background  D2O",
        background_action="add",
        bulk_in="Air",
        bulk_out="D2O",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="d13 acmw",
        data="d13acmw20",
        background="Background ACMW",
        background_action="add",
        bulk_in="Air",
        bulk_out="ACMW",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="d13 d2o",
        data="d13d2o20",
        background="Background  D2O",
        background_action="add",
        bulk_in="Air",
        bulk_out="D2O",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="d83 acmw",
        data="d83acmw20",
        background="Background ACMW",
        background_action="add",
        bulk_in="Air",
        bulk_out="ACMW",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="d83 d2o",
        data="d83d2o20",
        background="Background  D2O",
        background_action="add",
        bulk_in="Air",
        bulk_out="D2O",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )
    project.contrasts.append(
        name="fully h, D2O",
        data="hd2o20",
        background="Background  D2O",
        background_action="add",
        bulk_in="Air",
        bulk_out="D2O",
        scalefactor="Scalefactor 1",
        resolution="Resolution 1",
        resample=False,
        model=["Model_IIb"],
    )

    return project
